from dataclasses import dataclass, field
from functools import cached_property, cache
import itertools
import json
import networkx as nx
from typing import Dict, Set
from bddl.knowledge_base.orm import Model, ManyToOne, ManyToMany, OneToMany, ManyToOneField, ManyToManyField, OneToManyField, UUIDField
from bddl.knowledge_base.utils import *
from collections import defaultdict


ROOM_TYPE_CHOICES = [
    ('bar', 'bar'), 
    ('bathroom', 'bathroom'), 
    ('bedroom', 'bedroom'), 
    ('biology lab', 'biology_lab'), 
    ('break room', 'break_room'), 
    ('chemistry lab', 'chemistry_lab'),
    ('classroom', 'classroom'), 
    ('computer lab', 'computer_lab'), 
    ('conference hall', 'conference_hall'), 
    ('copy room', 'copy_room'), 
    ('corridor', 'corridor'), 
    ('dining room', 'dining_room'), 
    ('empty room', 'empty_room'), 
    ('grocery store', 'grocery_store'), 
    ('gym', 'gym'), 
    ('hammam', 'hammam'), 
    ('infirmary', 'infirmary'), 
    ('kitchen', 'kitchen'), 
    ('lobby', 'lobby'), 
    ('locker room', 'locker_room'), 
    ('meeting room', 'meeting_room'), 
    ('phone room', 'phone_room'), 
    ('private office', 'private_office'), 
    ('sauna', 'sauna'), 
    ('shared office', 'shared_office'), 
    ('spa', 'spa'), 
    ('entryway', 'entryway'), 
    ('television room', 'television_room'), 
    ('utility room', 'utility_room'), 
    ('garage', 'garage'), 
    ('closet', 'closet'), 
    ('childs room', 'childs_room'), 
    ('exercise room', 'exercise_room'), 
    ('garden', 'garden'), 
    ('living room', 'living_room'), 
    ('pantry room', 'pantry_room'), 
    ('playroom', 'playroom'), 
    ('staircase', 'staircase'), 
    ('storage room', 'storage_room')
]


@dataclass(eq=False, order=False)
class Property(Model):
    name : str
    parameters : str
    id : str = UUIDField()
    synset_fk : ManyToOne = ManyToOneField('Synset', 'properties')

    class Meta:
        pk = 'id'
        unique_together = ('synset', 'name')
        ordering = ['name']

    def __str__(self):
        return self.synset.name + '-' + self.name
        
    
@dataclass(eq=False, order=False)
class MetaLink(Model):
    name : str
    on_objects_fk : ManyToMany = ManyToManyField('Object', 'meta_links')

    class Meta:
        pk = 'name'


@dataclass(eq=False, order=False)
class Predicate(Model):
    name : str
    synsets_fk : ManyToMany = ManyToManyField('Synset', 'used_in_predicates')
    tasks_fk : ManyToMany = ManyToManyField('Task', 'uses_predicates')
    
    class Meta:
        pk = 'name'


@dataclass(eq=False, order=False)
class Scene(Model):
    name : str

    rooms_fk : OneToMany = OneToManyField('Room', 'scene')

    @cached_property
    def room_count(self):
        return len(self.rooms)
    
    @cached_property
    def object_count(self):
        return sum(
            roomobject.count
            for room in self.rooms
            for roomobject in room.roomobjects
            if not room.ready
        )
    
    @cached_property
    def any_ready(self):
        return any(room.ready for room in self.rooms)
    
    @cached_property
    def fully_ready(self):
        ready_count = sum(
            roomobject.count
            for room in self.rooms
            for roomobject in room.roomobjects
            if room.ready
        )
        unready_count = self.object_count
        return ready_count == unready_count
   
    class Meta:
        pk = 'name'
        ordering = ['name']


@dataclass(eq=False, order=False)
class Category(Model):
    name : str

    # the synset that the category belongs to
    synset_fk : ManyToOne = ManyToOneField('Synset', 'categories')

    # objects that belong to this category
    objects_fk : OneToMany = OneToManyField('Object', 'category')

    def __str__(self):
        return self.name
    
    class Meta:
        pk = 'name'
        ordering = ['name']

    @cache
    def matching_synset(self, synset) -> bool:
        return synset.name in self.matching_synsets

    @cached_property
    def matching_synsets(self) -> Set['Synset']:
        if not self.synset:
            return set()
        return {anc.name for anc in self.synset.ancestors} | {self.synset.name}

    @classmethod
    def view_mapped_to_substance_synset(cls):
        """Categories Incorrectly Mapped to Substance Synsets"""
        return [x for x in cls.all_objects() if x.category.synset.state == STATE_SUBSTANCE]
    
    @classmethod
    def view_mapped_to_non_leaf_synsets(cls):
        """Categories Mapped to Non-Leaf Synsets"""
        return [x for x in cls.all_objects() if len(x.synset.children) > 0]


@dataclass(eq=False, order=False)
class Object(Model):
    name : str
    # the name of the object prior to getting renamed
    original_name : str
    # providing target
    provider : str = ""
    # whether the object is in the current dataset
    ready : bool = False
    # whether the object is planned 
    planned : bool = True
    # the category that the object belongs to
    category_fk : ManyToOne = ManyToOneField(Category, 'objects')
    # meta links owned by the object
    meta_links_fk : ManyToMany = ManyToManyField(MetaLink, 'on_objects')
    # roomobject counts of this object
    roomobjects_fk : OneToMany = OneToManyField('RoomObject', 'object')

    def __str__(self):
        return self.name
    
    class Meta:
        pk = 'name'
        ordering = ['name']

    @cache
    def matching_synset(self, synset) -> bool:
        return self.category.matching_synset(synset)
    
    @cached_property
    def state(self):
        if self.ready:
            return STATE_MATCHED 
        elif self.planned:
            return STATE_PLANNED
        return STATE_UNMATCHED
    
    @cached_property
    def image_url(self):
        model_id = self.name.split('-')[-1]
        return f'https://cvgl.stanford.edu/b1k/object_images/{model_id}.webp'
    
    def fully_supports_synset(self, synset) -> bool:       
        return synset.required_meta_links.issubset({x.name for x in self.meta_links})
    
    @cached_property
    def missing_meta_links(self) -> List[str]:
        return sorted(self.category.synset.required_meta_links - {x.name for x in self.meta_links})

    @classmethod
    def view_objects_with_missing_meta_links(cls):
        """Objects with Missing Meta Links"""
        return [o for o in cls.all_objects() if not o.fully_supports_synset(o.category.synset)]


@dataclass(eq=False, order=False)
class Synset(Model):
    name : str
    # whether the synset is a custom synset or not
    is_custom : bool = field(default=False, repr=False)
    # wordnet definitions
    definition : str = field(default="", repr=False)
    # whether the synset is used as a substance in some task
    is_used_as_substance : bool = field(default=False, repr=False)
    # whether the synset is used as a non-substance in some task
    is_used_as_non_substance : bool = field(default=False, repr=False)
    # whether the synset is ever used as a fillable in any task
    is_used_as_fillable : bool = field(default=False, repr=False)
    # predicates the synset was used in as the first argument
    used_in_predicates_fk : ManyToMany = ManyToManyField(Predicate, 'synsets')
    # all it's parents in the synset graph (NOTE: this does not include self)
    parents_fk : ManyToMany = ManyToManyField('Synset', 'children')
    children_fk : ManyToMany = ManyToManyField('Synset', 'parents')
    # all ancestors (NOTE: this includes self)
    ancestors_fk : ManyToMany = ManyToManyField('Synset', 'descendants')
    descendants_fk : ManyToMany = ManyToManyField('Synset', 'ancestors')
    # state of the synset, one of STATE METADATA (pre computed to save webpage generation time)
    state : str = field(default=STATE_ILLEGAL, repr=False)

    categories_fk : OneToMany = OneToManyField(Category, 'synset')
    properties_fk : OneToMany = OneToManyField(Property, 'synset')
    tasks_fk : ManyToMany = ManyToManyField('Task', 'synsets')
    tasks_using_as_future_fk : ManyToMany = ManyToManyField('Task', 'future_synsets')
    used_by_transition_rules_fk : ManyToMany = ManyToManyField('TransitionRule', 'input_synsets')
    produced_by_transition_rules_fk : ManyToMany = ManyToManyField('TransitionRule', 'output_synsets')
    roomsynsetrequirements_fk : OneToMany = OneToManyField('RoomSynsetRequirement', 'synset')

    class Meta:
        pk = 'name'
        ordering = ['name']

    @cached_property
    def property_names(self):
        return {prop.name for prop in self.properties}

    @cached_property
    def direct_matching_objects(self) -> Set[Object]:
        matched_objs = set()
        for category in self.categories:
            matched_objs.update(category.objects)
        return matched_objs
    
    @cached_property
    def direct_matching_ready_objects(self) -> Set[Object]:
        matched_objs = set()
        for category in self.categories:
            matched_objs.update(x for x in category.objects if x.ready)
        return matched_objs

    @cached_property
    def matching_objects(self) -> Set[Object]:
        matched_objs = set(self.direct_matching_objects)
        for synset in self.descendants:
            matched_objs.update(synset.direct_matching_objects)
        return matched_objs
    
    @cached_property
    def matching_ready_objects(self) -> Set[Object]:
        matched_objs = set(self.direct_matching_ready_objects)
        for synset in self.descendants:
            matched_objs.update(synset.direct_matching_ready_objects)
        return matched_objs
    
    @cached_property
    def required_meta_links(self) -> Set[str]:
        properties = {prop.name: json.loads(prop.parameters) for prop in self.properties}

        if 'substance' in properties:
            return set()  # substances don't need any meta links
        
        required_links = set()

        # If we are a heatSource or coldSource, we need to have certain links
        for property in ['heatSource', 'coldSource']:
            if property in properties:
                if 'requires_inside' in properties[property] and properties[property]['requires_inside']:
                    continue
                required_links.add('heatsource')

        if 'fillable' in properties:
            required_links.add('fillable')

        if 'toggleable' in properties:
            required_links.add('togglebutton')

        particle_pairs = [
            ('particleSink', 'fluidsink'),
            ('particleSource', 'fluidsource'),
            ('particleApplier', 'particleapplier'),
            ('particleRemover', 'particleremover'),
        ]
        for property, meta_link in particle_pairs:
            if property in properties:
                if 'method' in properties[property] and properties[property]['method'] != 1: # not projection
                    continue
                required_links.add(meta_link)

        if 'slicer' in properties:
            required_links.add('slicer')

        if 'sliceable' in properties:
            required_links.add('subpart')

        return required_links
    
    @cached_property
    def has_fully_supporting_object(self) -> bool:
        if self.state == STATE_SUBSTANCE:
            return True

        for obj in self.matching_objects:
            if obj.fully_supports_synset(self):
                return True
        return False
    
    @cached_property
    def n_task_required(self):
        '''Get whether the synset is required in any task, returns STATE METADATA'''
        return len(self.tasks)
    
    @cached_property
    def subgraph(self):
        '''Get the edges of the subgraph of the synset'''
        G = nx.DiGraph()
        next_to_query = [(self, True, True)]
        while next_to_query:
            synset, query_parents, query_children = next_to_query.pop()
            if query_parents:
                for parent in synset.parents:
                    G.add_edge(parent, synset)
                    next_to_query.append((parent, True, False))
            if query_children:
                for child in synset.children:
                    G.add_edge(synset, child)
                    next_to_query.append((child, False, True))
        return G    

    @cached_property
    def task_relevant(self):
        return bool(self.tasks or any(ancestor.tasks for ancestor in self.ancestors))
    
    @cached_property
    def transition_subgraph(self):
        producing_recipes = list(self.produced_by_transition_rules)
        consuming_recipes = list(self.used_by_transition_rules)
        transitions = set(itertools.chain(producing_recipes, consuming_recipes))
        G = nx.DiGraph()
        for transition in transitions:
            G.add_node(transition)
            for input_synset in transition.input_synsets:
                G.add_edge(input_synset, transition)
            for output_synset in transition.output_synsets:
                G.add_edge(transition, output_synset)
        return nx.relabel_nodes(G, lambda x: (x.name if isinstance(x, Synset) else f'recipe: {x.name}'), copy=True)

    def is_produceable_from(self, synsets):
        def _is_produceable_from(_self, _synsets, _seen):
            if _self in _seen:
                return False, set()
            
            # If it's already available, then we're good.
            if _self in _synsets:
                return True, set()

            # Otherwise, are there any recipes that I can use to obtain it?
            recipe_alternatives = set()
            for recipe in _self.produced_by_transition_rules:
                producabilities_and_recipe_sets = [_is_produceable_from(ingredient, _synsets, _seen | {_self}) for ingredient in recipe.input_synsets]
                producabilities, recipe_sets = zip(*producabilities_and_recipe_sets)
                if all(producabilities):
                    recipe_alternatives.add(recipe)
                    recipe_alternatives.update(ingredient_recipe for recipe_set in recipe_sets for ingredient_recipe in recipe_set)
                
            if not recipe_alternatives:
                return False, set()
            
            return True, recipe_alternatives
        
        return _is_produceable_from(self, synsets, set())

    @cached_property
    def is_derivative(self):
        derivative_words = ["cooked__", "half__", "diced__"]
        return any(self.name.startswith(dw) for dw in derivative_words)
        
    @cached_property
    def derivative_parent(self):
        # Check if the synset is a derivative
        if not self.is_derivative:
            return None

        # Get the parent name
        parent_name = self.name.split(".n.")[0].split("__", 1)[-1]
        if self.name.startswith("diced__"):
            parent_name = "half__" + parent_name

        # Otherwise make a set of candidates
        parent_should_be_substance = self.name.startswith("cooked__")
        parent_should_have_properties = set()

        if self.name.startswith("diced__") or self.name.startswith("half__"):
            parent_should_have_properties.add("sliceable")
        elif self.name.startswith("cooked__"):
            parent_should_have_properties.add("cookable")

        parent_candidates = [
            s for s in Synset.all_objects()
            if (s.state == STATE_SUBSTANCE) == parent_should_be_substance and
            s.name.split(".n.")[0] == parent_name and
            len(parent_should_have_properties - s.property_names) == 0
        ]

        assert len(parent_candidates) <= 1, f"Multiple candidates for parent of {self.name}: {parent_candidates}"

        # Return the parent if it exists
        return parent_candidates[0] if parent_candidates else None
    
    @cached_property
    def derivative_root(self):
        if not self.derivative_parent:
            return None
        
        parent_root = self.derivative_parent.derivative_root
        return parent_root if parent_root else self.derivative_parent
    
    @cached_property
    def derivative_children(self):
        return {s for s in Synset.all_objects() if s.derivative_parent == self}
    
    @cached_property
    def derivative_ancestors(self):
        if not self.derivative_parent:
            return {self}
        return {self, self.derivative_parent} | set(self.derivative_parent.derivative_ancestors)

    @cached_property
    def derivative_descendants(self):
        descendants = {self}.update(self.derivative_children)
        for child in self.derivative_children:
            descendants.update(child.derivative_descendants)
        return descendants

    @classmethod
    def view_substance_mismatch(cls):
        """Synsets that are used in predicates that don't match their substance state"""
        return [
            s for s in cls.all_objects()
            if (
                (s.state == STATE_SUBSTANCE and s.is_used_as_non_substance) or 
                (not s.state == STATE_SUBSTANCE and s.is_used_as_substance) or 
                (s.is_used_as_substance and s.is_used_as_non_substance)
            )]

    @classmethod
    def view_object_unsupported_properties(cls):
        """Leaf synsets that do not have at least one object that supports all of annotated properties."""
        # TODO: joint count
        return [
            s for s in cls.all_objects()
            if len(s.matching_objects) > 0 and len(s.children) == 0 and not s.has_fully_supporting_object
        ]
    
    @classmethod
    def view_unnecessary(cls):
        """Objectless synsets that are not used in any task or required by any property"""
        transition_relevant_synsets = {
            anc
            for t in Task.all_objects()
            for transition in t.relevant_transitions
            for s in list(transition.output_synsets) + list(transition.input_synsets)
            for anc in s.ancestors
        }
        return [
            s for s in cls.all_objects()
            if all(
                not sp.task_relevant and
                sp not in transition_relevant_synsets and
                len(sp.matching_objects) == 0 and
                len(sp.children) == 0
                for sp in s.derivative_ancestors
            )
        ]
    
    @classmethod
    def view_bad_derivative(cls):
        """Derivative synsets that exist even though the original synset is missing the expected property"""
        return [s for s in cls.all_objects() if s.is_derivative and not s.derivative_parent]
    
    @classmethod
    def view_missing_derivative(cls):
        """Synsets that are missing a derivative synset that is expected to exist from property annotations"""
        # TODO: reimplement using new derivative fields
        sliceables = [
            s for s in cls.all_objects()
            if "sliceable" in s.property_names and s.state != STATE_SUBSTANCE
        ]
        missing_half = [
            s for s in sliceables
            if not cls.get("half__" + s.name.split(".n.")[0] + ".n.01")
        ]
        missing_diced = [
            s for s in sliceables
            if not cls.get("diced__" + s.name.split(".n.")[0] + ".n.01")
        ]

        cookable_substances = [
            s for s in cls.all_objects()
            if "cookable" in s.property_names and s.state == STATE_SUBSTANCE
        ]
        missing_cooked = [
            s for s in cookable_substances
            if not cls.get("cooked__" + s.name)
        ]

        return sorted(missing_half + missing_diced + missing_cooked)

@dataclass(eq=False, order=False)
class TransitionRule(Model):
    name : str
    input_synsets_fk : ManyToMany = ManyToManyField(Synset, 'used_by_transition_rules')
    output_synsets_fk : ManyToMany = ManyToManyField(Synset, 'produced_by_transition_rules')

    class Meta:
        pk = 'name'
        ordering = ['name']

    @cached_property
    def subgraph(self):
        G = nx.DiGraph()
        G.add_node(self)
        for input_synset in self.input_synsets:
            G.add_edge(input_synset, self)
        for output_synset in self.output_synsets:
            G.add_edge(self, output_synset)
        return nx.relabel_nodes(G, lambda x: (x.name if isinstance(x, Synset) else f'recipe: {x.name}'), copy=True)

    @staticmethod
    def get_graph():
        G = nx.DiGraph()
        for transition in TransitionRule.all_objects():
            G.add_node(transition)
            for input_synset in transition.input_synsets:
                G.add_edge(input_synset, transition)
            for output_synset in transition.output_synsets:
                G.add_edge(transition, output_synset)
        return G


@dataclass(eq=False, order=False)
class Task(Model):
    name : str
    definition : str
    synsets_fk : ManyToMany = ManyToManyField(Synset, 'tasks') # the synsets required by this task
    future_synsets_fk : ManyToMany = ManyToManyField(Synset, 'tasks_using_as_future') # the synsets that show up as future synsets in this task (e.g. don't exist in initial)
    uses_predicates_fk : ManyToMany = ManyToManyField(Predicate, 'tasks')
    room_requirements_fk : OneToMany = OneToManyField('RoomRequirement', 'task')
    
    class Meta:
        pk = 'name'
        ordering = ['name']

    @cached_property
    def state(self):
        if self.synset_state == STATE_MATCHED and self.scene_state == STATE_MATCHED:
            return STATE_MATCHED
        elif self.synset_state == STATE_UNMATCHED or self.scene_state == STATE_UNMATCHED:
            return STATE_UNMATCHED
        else:
            return STATE_PLANNED

    def matching_scene(self, scene: Scene, ready: bool=True) -> str:
        '''checks whether a scene satisfies task requirements'''
        ret = ''
        for room_requirement in self.room_requirements:
            scene_ret = f'Cannot find suitable {room_requirement.type}: '
            for room in scene.rooms:
                if room.type != room_requirement.type or room.ready != ready:
                    continue
                room_ret = room.matching_room_requirement(room_requirement)
                if len(room_ret)== 0:
                    scene_ret = ''
                    break
                else: 
                    scene_ret += f'{room.name} is missing {room_ret}; '
            if len(scene_ret) > 0:
                ret += scene_ret[:-2] + '.'
        return ret
    
    def uses_transition(self):
        return any(pred.name == "future" for pred in self.uses_predicates)
    
    def uses_visual_substance(self):
        return any("visualSubstance" in synset.property_names for synset in self.synsets)

    def uses_physical_substance(self):
        return any("physicalSubstance" in synset.property_names for synset in self.synsets)

    def uses_attachment(self):
        return any(pred.name in ['assembled', 'attached'] for pred in self.uses_predicates)

    def uses_cloth(self):
        return any(pred.name in ['folded', 'draped', 'unfolded'] for pred in self.uses_predicates)

    @cached_property
    def substance_synsets(self):
        '''synsets that represent a substance'''
        return [x for x in self.synsets if x.state == STATE_SUBSTANCE]
    
    @cached_property
    def synset_state(self) -> str:
        if any(synset.state == STATE_ILLEGAL for synset in self.synsets):
            return STATE_UNMATCHED
        elif any(synset.state == STATE_UNMATCHED for synset in self.synsets):
            return STATE_UNMATCHED
        elif any(synset.state == STATE_PLANNED for synset in self.synsets):
            return STATE_PLANNED
        else:
            return STATE_MATCHED
        
    @cached_property
    def problem_synsets(self):
        return [synset for synset in self.synsets if synset.state in (STATE_ILLEGAL, STATE_UNMATCHED)]
       
    @cached_property
    def scene_matching_dict(self) -> Dict[str, Dict[str, str]]:
        ret = {}
        for scene in Scene.all_objects():
            if not any(room.ready for room in scene.rooms):
                result_ready = 'Scene does not have a ready version currently.'
            else:
                result_ready = self.matching_scene(scene=scene, ready=True)
            result_partial = self.matching_scene(scene=scene, ready=False)
            ret[scene] = {
                'matched_ready': len(result_ready) == 0,
                'reason_ready': result_ready,
                'matched_planned': len(result_partial) == 0,
                'reason_planned': result_partial,
            }
        return ret
    
    @cached_property
    def scene_state(self) -> str:
        scene_matching_dict = self.scene_matching_dict
        if any(x['matched_ready'] for x in scene_matching_dict.values()):
            return STATE_MATCHED
        elif any(x['matched_planned'] for x in scene_matching_dict.values()):
            return STATE_PLANNED
        else:
            return STATE_UNMATCHED
        
    @cached_property
    def substance_required(self) -> str:
        if self.substance_synsets:
            return STATE_SUBSTANCE
        else:
            return STATE_NONE
        
    @cached_property
    def producability_data(self) -> Dict[Synset, Tuple[bool, Set[TransitionRule]]]:
        '''Map each synset to a tuple of whether it is produceable and the transition rules that can be used to produce it'''
        starting_synsets = set(self.synsets) - set(self.future_synsets)
        return {synset: synset.is_produceable_from(starting_synsets) for synset in self.synsets}
    
    @cached_property
    def relevant_transitions(self):
        return [rule for synset, (producability, rules) in self.producability_data.items() for rule in rules]
    
    @cached_property
    def transition_graph(self):
        G  = nx.DiGraph()
        future_synsets = set(self.future_synsets)
        starting_synsets = set(self.synsets) - future_synsets
        
        def human_readable_name(s):
            if s in starting_synsets:
                return f'initial: {s.name}'
            elif s in future_synsets:
                return f'future: {s.name}'
            else:
                return s.name

        for synset in self.synsets:
            G.add_node(human_readable_name(synset), type='obj')
        for transition in self.relevant_transitions:
            transition_name = f'recipe: {transition.name}'
            G.add_node(transition_name, type='transition', text=transition.name)
            for input_synset in transition.input_synsets:
                G.add_edge(human_readable_name(input_synset), transition_name)
            for output_synset in transition.output_synsets:
                G.add_edge(transition_name, human_readable_name(output_synset))

        # Prune the graph so that it only contains all the future nodes and everything that can be used
        # to transition into them.
        future_nodes = {human_readable_name(s) for s in future_synsets}
        future_node_trees = [nx.bfs_tree(G.reverse(), node) for node in future_nodes]
        future_node_tree_nodes = [node for tree in future_node_trees for node in tree.nodes]
        subgraph = G.subgraph(future_node_tree_nodes)

        return subgraph
    
    @cached_property
    def partial_transition_graph(self):
        future_synsets = set(self.future_synsets)
        starting_synsets = set(self.synsets) - future_synsets

        def human_readable_name(s):
            if isinstance(s, Synset):
                if s in starting_synsets:
                    return f'initial: {s.name}'
                elif s in future_synsets:
                    return f'future: {s.name}'
                else:
                    return f'missing: {s.name}'
            elif isinstance(s, TransitionRule):
                return f'recipe: {s.name}'
            else:
                raise ValueError('Unexpected node.')
            
        # Build a graph from all the transitions
        G = TransitionRule.get_graph()
            
        # Show all the nodes that show up in some path from the initial to the future synsets.
        all_nodes_on_path = set()
        for f, t in itertools.product(starting_synsets, future_synsets):
            if f not in G or t not in G:
                continue
            for p in nx.all_simple_paths(G, f, t):
                all_nodes_on_path.update(p)

        # Also add all ingredients of each recipe
        all_recipes = [x for x in all_nodes_on_path if isinstance(x, TransitionRule)]
        all_ingredients = {inp for recipe in all_recipes for inp in recipe.input_synsets}
        all_nodes_to_keep = all_nodes_on_path | all_ingredients

        # Get the subgraph, convert that to a text graph.
        subgraph = G.subgraph(all_nodes_to_keep)
        return nx.relabel_nodes(subgraph, human_readable_name, copy=True)
    
    @cached_property
    def unreachable_goal_synsets(self) -> List[str]:
        '''Get a list of synsets that are in the goal but cannot be reached from the initial state'''
        return [s for s in self.future_synsets if not self.producability_data[s][0]]

    @cached_property
    def goal_is_reachable(self) -> bool:
        '''Get whether the goal is reachable from the initial state'''
        return len(self.unreachable_goal_synsets) == 0

    @classmethod
    def view_transition_failure(cls):
        """Transition Failure Tasks"""
        return [x for x in cls.all_objects() if not x.goal_is_reachable]
    
    @classmethod
    def view_non_scene_matched(cls):
        """Non-Scene-Matched Tasks"""
        return [x for x in cls.all_objects() if x.scene_state == STATE_UNMATCHED]


@dataclass(eq=False, order=False)
class RoomRequirement(Model):
    # TODO: make this one of the room types. enum?
    type : str
    id : str = UUIDField()
    task_fk : ManyToOne = ManyToOneField(Task, 'room_requirements')
    roomsynsetrequirements_fk : OneToMany = OneToManyField('RoomSynsetRequirement', 'room_requirement')

    class Meta:
        pk = 'id'
        unique_together = ('task', 'type')
        ordering = ['type']  


@dataclass(eq=False, order=False)
class RoomSynsetRequirement(Model):
    id : str = UUIDField()
    room_requirement_fk : ManyToOne = ManyToOneField(RoomRequirement, 'roomsynsetrequirements')
    synset_fk : ManyToOne = ManyToOneField(Synset, 'roomsynsetrequirements')
    count : int = 0

    class Meta:
        pk = 'id'
        unique_together = ('room_requirement', 'synset')
        ordering = ['synset__name']
    

@dataclass(eq=False, order=False)
class Room(Model):
    name : str
    # type of the room
    # TODO: make this one of the room types
    type : str
    # whether the scene is ready in the current dataset
    ready : bool = False
    id : str = UUIDField()
    # the scene the room belongs to
    scene_fk : ManyToOne = ManyToOneField(Scene, 'rooms')
    # the room objects in this object
    roomobjects_fk : OneToMany = OneToManyField('RoomObject', 'room')

    class Meta:
        pk = 'id'
        unique_together = ('name', 'ready', 'scene')
        ordering = ['name']
    
    def __str__(self):
        return f'{self.scene.name}_{self.type}_{"ready" if self.ready else "planned"}' 

    def matching_room_requirement(self, room_requirement: RoomRequirement) -> str:
        '''
        checks whether the room satisfies the room requirement from a task
        returns an empty string if it does, otherwise returns a string describing what is missing
        '''
        G = nx.Graph()
        # Add a node for each required object
        synset_node_to_synset = {}
        for room_synset_requirement in room_requirement.roomsynsetrequirements:
            synset_name = room_synset_requirement.synset.name
            for i in range(room_synset_requirement.count):
                node_name = f'{synset_name}_{i}'
                G.add_node(node_name)
                synset_node_to_synset[node_name] = room_synset_requirement.synset
        # Add a node for each object in the room
        for roomobject in self.roomobjects:
            for i in range(roomobject.count):
                object_name = f'{roomobject.object.name}_{i}'
                G.add_node(object_name)
                # Add edges to all matching synsets
                for synset_node, synset in synset_node_to_synset.items():
                    if roomobject.object.matching_synset(synset):
                        G.add_edge(object_name, synset_node)
        # Now do a bipartite matching
        M = nx.bipartite.maximum_matching(G, top_nodes=synset_node_to_synset.keys())
        # Now check that all required objects are matched
        if set(synset_node_to_synset.keys()).issubset(M.keys()):
            return ''
        else: 
            missing_synsets = defaultdict(int)  # default value is 0
            for synset_node, synset in synset_node_to_synset.items():
                if synset_node not in M:
                    missing_synsets[synset.name] += 1
            return ', '.join([f'{count} {synset}' for synset, count in missing_synsets.items()])


@dataclass(eq=False, order=False)
class RoomObject(Model):
    id : str = UUIDField()
    # the room that the object belongs to
    room_fk : ManyToOne = ManyToOneField(Room, 'roomobjects')
    # the actual object that the room object maps to
    object_fk : ManyToOne = ManyToOneField(Object, 'roomobjects')
    # number of objects in the room
    count : int = 0

    class Meta:
        pk = 'id'
        unique_together = ('room', 'object')
        ordering = ['room__name', 'object__name']

    def __str__(self):
        return f'{str(self.room)}_{self.object.name}'   



import copy
import itertools

import numpy as np

import bddl
from bddl.logic_base import Expression
from bddl.utils import (
    UnsupportedPredicateError,
    truncated_permutations,
    truncated_product,
)

#################### RECURSIVE PREDICATES ####################

# -JUNCTIONS


class Conjunction(Expression):
    def __init__(self, scope, backend, body, object_map, generate_ground_options=True):
        super().__init__(scope, backend, body, object_map)

        new_scope = copy.copy(scope)
        child_predicates = [
            get_predicate_for_token(subexpression[0], backend)(
                scope, 
                backend, 
                subexpression[1:],
                object_map,
                generate_ground_options=generate_ground_options
            )
            for subexpression in body
        ]
        self.children.extend(child_predicates)

        if generate_ground_options:
            self.get_ground_options()

    def evaluate(self):
        self.child_values = [child.evaluate() for child in self.children]
        assert all([val is not None for val in self.child_values]), "child_values has NoneTypes"
        return all(self.child_values)

    def get_ground_options(self):
        options = list(itertools.product(*[child.flattened_condition_options for child in self.children]))
        self.flattened_condition_options = []
        for option in options:
            self.flattened_condition_options.append(list(itertools.chain(*option)))


class Disjunction(Expression):
    def __init__(self, scope, backend, body, object_map, generate_ground_options=True):
        super().__init__(scope, backend, body, object_map)

        # body = [[predicate1], [predicate2], ..., [predicateN]]
        new_scope = copy.copy(scope)
        child_predicates = [
            get_predicate_for_token(subexpression[0], backend)(
                scope, 
                backend, 
                subexpression[1:],
                object_map,
                generate_ground_options=generate_ground_options
            )
            for subexpression in body
        ]
        self.children.extend(child_predicates)

        if generate_ground_options:
            self.get_ground_options()

    def evaluate(self):
        self.child_values = [child.evaluate() for child in self.children]
        assert all([val is not None for val in self.child_values]), "child_values has NoneTypes"
        return any(self.child_values)

    def get_ground_options(self):
        self.flattened_condition_options = []
        for child in self.children:
            self.flattened_condition_options.extend(child.flattened_condition_options)


# QUANTIFIERS
class Universal(Expression):
    def __init__(self, scope, backend, body, object_map, generate_ground_options=True):
        super().__init__(scope, backend, body, object_map)
        iterable, subexpression = body
        param_label, __, category = iterable
        param_label = param_label.strip("?")
        assert __ == "-", "Middle was not a hyphen"
        for obj_name, obj in scope.items():
            if obj_name in object_map[category]:
                new_scope = copy.copy(scope)
                new_scope[param_label] = obj_name
                self.children.append(
                    get_predicate_for_token(subexpression[0], backend)(
                        new_scope, 
                        backend, 
                        subexpression[1:],
                        object_map,
                        generate_ground_options=generate_ground_options
                    )
                )

        if generate_ground_options:
            self.get_ground_options()
            
    def evaluate(self):
        self.child_values = [child.evaluate() for child in self.children]
        assert all([val is not None for val in self.child_values]), "child_values has NoneTypes"
        return all(self.child_values)

    def get_ground_options(self):
        # Accept just a few possible options
        options = list(truncated_product(*[child.flattened_condition_options for child in self.children]))
        self.flattened_condition_options = []
        for option in options:
            self.flattened_condition_options.append(list(itertools.chain(*option)))


class Existential(Expression):
    def __init__(self, scope, backend, body, object_map, generate_ground_options=True):
        super().__init__(scope, backend, body, object_map)
        iterable, subexpression = body
        param_label, __, category = iterable
        param_label = param_label.strip("?")
        assert __ == "-", "Middle was not a hyphen"
        for obj_name, obj in scope.items():
            if obj_name in object_map[category]:
                new_scope = copy.copy(scope)
                new_scope[param_label] = obj_name
                # body = [["param_label", "-", "category"], [predicate]]
                self.children.append(
                    get_predicate_for_token(subexpression[0], backend)(
                        new_scope, 
                        backend, 
                        subexpression[1:],
                        object_map,
                        generate_ground_options=generate_ground_options
                    )
                )

        if generate_ground_options:
            self.get_ground_options()
            
    def evaluate(self):
        self.child_values = [child.evaluate() for child in self.children]
        assert all([val is not None for val in self.child_values]), "child_values has NoneTypes"
        return any(self.child_values)

    def get_ground_options(self):
        self.flattened_condition_options = []
        for child in self.children:
            self.flattened_condition_options.extend(child.flattened_condition_options)


class NQuantifier(Expression):
    def __init__(self, scope, backend, body, object_map, generate_ground_options=True):
        super().__init__(scope, backend, body, object_map)

        N, iterable, subexpression = body
        self.N = int(N[0])
        param_label, __, category = iterable
        param_label = param_label.strip("?")
        assert __ == "-", "Middle was not a hyphen"
        for obj_name, obj in scope.items():
            if obj_name in object_map[category]:
                new_scope = copy.copy(scope)
                new_scope[param_label] = obj_name
                self.children.append(
                    get_predicate_for_token(subexpression[0], backend)(
                        new_scope, 
                        backend, 
                        subexpression[1:],
                        object_map,
                        generate_ground_options=generate_ground_options
                    )
                )

        if generate_ground_options:
            self.get_ground_options()
            
    def evaluate(self):
        self.child_values = [child.evaluate() for child in self.children]
        assert all([val is not None for val in self.child_values]), "child_values has NoneTypes"
        return sum(self.child_values) == self.N

    def get_ground_options(self):
        # Accept just a few possible options
        options = list(truncated_product(*[child.flattened_condition_options for child in self.children]))
        self.flattened_condition_options = []
        for option in options:
            # for combination in [combo for num_el in range(self.N - 1, len(option)) for combo in itertools.combinations(option, num_el + 1)]:
            # Use a minimal solution (exactly N fulfilled, rather than >=N fulfilled)
            for combination in itertools.combinations(option, self.N):
                self.flattened_condition_options.append(list(itertools.chain(*combination)))


class ForPairs(Expression):
    def __init__(self, scope, backend, body, object_map, generate_ground_options=True):
        super().__init__(scope, backend, body, object_map)

        iterable1, iterable2, subexpression = body
        param_label1, __, category1 = iterable1
        param_label2, __, category2 = iterable2
        param_label1 = param_label1.strip("?")
        param_label2 = param_label2.strip("?")
        for obj_name_1, obj_1 in scope.items():
            if obj_name_1 in object_map[category1]:
                sub = []
                for obj_name_2, obj_2 in scope.items():
                    if obj_name_2 in object_map[category2] and obj_name_1 != obj_name_2:
                        new_scope = copy.copy(scope)
                        new_scope[param_label1] = obj_name_1
                        new_scope[param_label2] = obj_name_2
                        sub.append(
                            get_predicate_for_token(subexpression[0], backend)(
                                new_scope, 
                                backend, 
                                subexpression[1:],
                                object_map,
                                generate_ground_options=generate_ground_options
                            )
                        )
                self.children.append(sub)

        if generate_ground_options:
            self.get_ground_options()
            
    def evaluate(self):
        self.child_values = np.array([np.array([subchild.evaluate() for subchild in child]) for child in self.children])

        L = min(len(self.children), len(self.children[0]))
        return (np.sum(np.any(self.child_values, axis=1), axis=0) >= L) and (
            np.sum(np.any(self.child_values, axis=0), axis=0) >= L
        )

    def get_ground_options(self):
        self.flattened_condition_options = []
        M, N = len(self.children), len(self.children[0])
        L, G = min(M, N), max(M, N)
        all_L_choices = truncated_permutations(range(L))
        all_G_choices = truncated_permutations(range(G), r=L)
        for lchoice in all_L_choices:
            for gchoice in all_G_choices:
                if M < N:
                    all_child_options = [
                        self.children[lchoice[l]][gchoice[l]].flattened_condition_options for l in range(L)
                    ]
                else:
                    all_child_options = [
                        self.children[gchoice[l]][lchoice[l]].flattened_condition_options for l in range(L)
                    ]
                choice_options = truncated_product(*all_child_options)
                unpacked_choice_options = []
                for choice_option in choice_options:
                    unpacked_choice_options.append(list(itertools.chain(*choice_option)))
                self.flattened_condition_options.extend(unpacked_choice_options)


class ForNPairs(Expression):
    def __init__(self, scope, backend, body, object_map, generate_ground_options=True):
        super().__init__(scope, backend, body, object_map)

        N, iterable1, iterable2, subexpression = body
        self.N = int(N[0])
        param_label1, __, category1 = iterable1
        param_label2, __, category2 = iterable2
        param_label1 = param_label1.strip("?")
        param_label2 = param_label2.strip("?")
        for obj_name_1, obj_1 in scope.items():
            if obj_name_1 in object_map[category1]:
                sub = []
                for obj_name_2, obj_2 in scope.items():
                    if obj_name_2 in object_map[category2] and obj_name_1 != obj_name_2:
                        new_scope = copy.copy(scope)
                        new_scope[param_label1] = obj_name_1
                        new_scope[param_label2] = obj_name_2
                        sub.append(
                            get_predicate_for_token(subexpression[0], backend)(
                                new_scope, 
                                backend, 
                                subexpression[1:],
                                object_map,
                                generate_ground_options=generate_ground_options
                            )
                        )
                self.children.append(sub)

        if generate_ground_options:
            self.get_ground_options()
            
    def evaluate(self):
        self.child_values = np.array([np.array([subchild.evaluate() for subchild in child]) for child in self.children])
        return (np.sum(np.any(self.child_values, axis=1), axis=0) >= self.N) and (
            np.sum(np.any(self.child_values, axis=0), axis=0) >= self.N
        )

    def get_ground_options(self):
        self.flattened_condition_options = []
        P, Q = len(self.children), len(self.children[0])
        L = min(P, Q)
        assert self.N <= L, "ForNPairs asks for more pairs than instances available"
        all_P_choices = truncated_permutations(range(P), r=self.N)
        all_Q_choices = truncated_permutations(range(Q), r=self.N)
        for pchoice in all_P_choices:
            for qchoice in all_Q_choices:
                all_child_options = [
                    self.children[pchoice[n]][qchoice[n]].flattened_condition_options for n in range(self.N)
                ]
                choice_options = truncated_product(*all_child_options)
                unpacked_choice_options = []
                for choice_option in choice_options:
                    unpacked_choice_options.append(list(itertools.chain(*choice_option)))
                self.flattened_condition_options.extend(unpacked_choice_options)


# NEGATION
class Negation(Expression):
    def __init__(self, scope, backend, body, object_map, generate_ground_options=True):
        super().__init__(scope, backend, body, object_map)

        # body = [[predicate]]
        subexpression = body[0]
        self.children.append(
            get_predicate_for_token(subexpression[0], backend)(
                scope, 
                backend, 
                subexpression[1:],
                object_map,
                generate_ground_options=generate_ground_options
            )
        )
        assert len(self.children) == 1, "More than one child."

        if generate_ground_options:
            self.get_ground_options()
            
    def evaluate(self):
        self.child_values = [child.evaluate() for child in self.children]
        assert len(self.child_values) == 1, "More than one child value"
        assert all([val is not None for val in self.child_values]), "child_values has NoneTypes"
        return not self.child_values[0]

    def get_ground_options(self):
        # demorgan's law
        self.flattened_condition_options = []
        child = self.children[0]
        negated_options = []
        for option in child.flattened_condition_options:
            negated_conds = []
            for cond in option:
                negated_conds.append(["not", cond])
            negated_options.append(negated_conds)
        # only picking one condition from each set of disjuncts
        for negated_option_selections in itertools.product(*negated_options):
            self.flattened_condition_options.append(list(itertools.chain(negated_option_selections)))


# IMPLICATION
class Implication(Expression):
    def __init__(self, scope, backend, body, object_map, generate_ground_options=True):
        super().__init__(scope, backend, body, object_map)

        # body = [[antecedent], [consequent]]
        antecedent, consequent = body
        self.children.append(
            get_predicate_for_token(antecedent[0], backend)(
                scope, 
                backend, 
                antecedent[1:],
                object_map,
                generate_ground_options=generate_ground_options
            )
        )
        self.children.append(
            get_predicate_for_token(consequent[0], backend)(
                scope, 
                backend, 
                consequent[1:],
                object_map,
                generate_ground_options=generate_ground_options
            )
        )

        if generate_ground_options:
            self.get_ground_options()
            
    def evaluate(self):
        self.child_values = [child.evaluate() for child in self.children]
        assert all([val is not None for val in self.child_values]), "child_values has NoneTypes"
        ante, cons = self.child_values
        return (not ante) or cons

    def get_ground_options(self):
        # (not antecedent) or consequent
        flattened_neg_antecedent_options = []
        antecedent = self.children[0]
        negated_options = []
        for option in antecedent.flattened_condition_options:
            negated_conds = []
            for cond in option:
                negated_conds.append(["not", cond])
            negated_options.append(negated_conds)
        for negated_option_selections in itertools.product(*negated_options):
            flattened_neg_antecedent_options.append(list(itertools.chain(negated_option_selections)))

        flattened_consequent_options = self.children[1].flattened_condition_options

        self.flattened_condition_options = flattened_neg_antecedent_options + flattened_consequent_options


# HEAD


class HEAD(Expression):
    def __init__(self, scope, backend, body, object_map, generate_ground_options=True):
        super().__init__(scope, backend, body, object_map)

        subexpression = body
        self.children.append(
            get_predicate_for_token(subexpression[0], backend)(
                scope, 
                backend, 
                subexpression[1:],
                object_map,
                generate_ground_options=generate_ground_options
            )
        )

        self.terms = [term.lstrip("?") for term in list(flatten_list(self.body))]

        if generate_ground_options:
            self.get_ground_options()
            
    def evaluate(self):
        self.child_values = [child.evaluate() for child in self.children]
        assert len(self.child_values) == 1, "More than one child value"
        self.currently_satisfied = self.child_values[0]
        return self.currently_satisfied

    def get_relevant_objects(self):
        # All object instances and categories that are in the scope will be collected
        objects = set([self.scope[obj_name] for obj_name in self.terms if obj_name in self.scope])

        # If this has a quantifier, the category-relevant objects won't all be caught, so adding them here
        # No matter what the quantifier, every object of the category/ies is relevant
        for term in self.terms:
            if term in self.object_map:
                for obj_name, obj in self.scope.items():
                    if obj_name in self.object_map[term]:
                        objects.add(obj)

        return list(objects)

    def get_ground_options(self):
        self.flattened_condition_options = self.children[0].flattened_condition_options


#################### CHECKING ####################


def create_scope(object_terms):
    """
    Creates degenerate scope mapping all object parameters to None
    :param objects: (list of strings) BDDL terms for objects
    """
    scope = {}
    for object_cat in object_terms:
        for object_inst in object_terms[object_cat]:
            scope[object_inst] = None
    return scope


def compile_state(parsed_state, backend, scope=None, object_map=None, generate_ground_options=True):
    compiled_state = []
    for parsed_condition in parsed_state:
        scope = scope if scope is not None else {}
        compiled_state.append(HEAD(
            scope, 
            backend, 
            parsed_condition, 
            object_map, 
            generate_ground_options=generate_ground_options
        ))
    return compiled_state

def evaluate_state(compiled_state):
    results = {"satisfied": [], "unsatisfied": []}
    for i, compiled_condition in enumerate(compiled_state):
        if compiled_condition.evaluate():
            results["satisfied"].append(i)
        else:
            results["unsatisfied"].append(i)
    return not bool(results["unsatisfied"]), results


def get_ground_state_options(compiled_state, backend, scope=None, object_map=None):
    all_options = list(
        itertools.product(*[compiled_condition.flattened_condition_options for compiled_condition in compiled_state])
    )
    all_unpacked_options = [list(itertools.chain(*option)) for option in all_options]

    # Remove all unsatisfiable options (those that contain some (cond1 and not cond1))
    consistent_unpacked_options = []
    for option in all_unpacked_options:
        consistent = True
        for cond1, cond2 in itertools.combinations(option, 2):
            if (cond1[0] == "not" and cond1[1] == cond2) or (cond2[0] == "not" and cond2[1] == cond1):
                consistent = False
                break
        if not consistent:
            continue
        consistent_unpacked_options.append(option)

    consistent_unpacked_options = [
        compile_state(option, backend, scope=scope, object_map=object_map)
        for option in sorted(consistent_unpacked_options, key=len)
    ]
    return consistent_unpacked_options


#################### UTIL ######################


def flatten_list(li):
    for elem in li:
        if isinstance(elem, list):
            yield from flatten_list(elem)
        else:
            yield elem


#################### TOKEN MAPPING ####################

TOKEN_MAPPING = {
    # BDDL
    "forall": Universal,
    "exists": Existential,
    "and": Conjunction,
    "or": Disjunction,
    "not": Negation,
    "imply": Implication,
    # BDDL extensions
    "forn": NQuantifier,
    "forpairs": ForPairs,
    "fornpairs": ForNPairs,
}


def get_predicate_for_token(token, backend):
    if token in TOKEN_MAPPING:
        return TOKEN_MAPPING[token]
    else:
        try:
            return backend.get_predicate_class(token)
        except KeyError as e:
            raise UnsupportedPredicateError(e)



import pathlib
import json
import os
from collections.abc import Iterable

TRANSITION_RULE_FOLDER = pathlib.Path(__file__).parents[1] / "generated_data" / "transition_map" / "tm_jsons"
SYNSET_KEYS = ["machine", "container", "washed_item", "heat_source", "input_synsets", "output_synsets"]
MACHINE_KEY = ["machine"]
CONTAINER_KEY = ["container", "machine"]

def sanity_check_object_hierarchy(object_taxonomy):
    leaf_synsets = object_taxonomy.get_leaf_descendants("entity.n.01")
    leaf_substance_synsets = {synset for synset in leaf_synsets if object_taxonomy.has_ability(synset, "substance")}
    for s in leaf_substance_synsets:
        substances = object_taxonomy.get_substances(s)
        assert len(substances) == 1, f"Substance synset {s} is mapped to zero or more than one substance: {substances}."

    for s in leaf_synsets:
        abilities = object_taxonomy.get_abilities(s)
        for ability in ["particleApplier", "particleRemover", "particleSource", "particleSink"]:
            if ability in abilities:
                ability_params = abilities[ability]
                for substance_synset in ability_params["conditions"]:
                    assert substance_synset in leaf_substance_synsets, f"In ParticleModifier annotation, {substance_synset} is not a leaf substance synset."
                    conditions = ability_params["conditions"][substance_synset]
                    if conditions is not None:
                        for condition in conditions:
                            if condition[0] == "saturated":
                                assert condition[1] in leaf_substance_synsets, f"In ParticleModifier annotation, {condition[1]} is not a leaf substance synset."

def sanity_check_transition_rules(object_taxonomy):
    leaf_synsets = object_taxonomy.get_leaf_descendants("entity.n.01")
    valid_keys = set()
    for json_file in os.listdir(TRANSITION_RULE_FOLDER):
        with open(TRANSITION_RULE_FOLDER / json_file, "r") as f:
            transition_rule = json.load(f)
            for rule in transition_rule:
                for key in rule:
                    if key in SYNSET_KEYS:
                        val = rule[key]
                        for s in val:
                            assert s in leaf_synsets, f"In transition rule file {json_file}, rule {rule}, {s} is not a leaf synset."

                    if key in MACHINE_KEY:
                        val = rule[key]
                        for s in val:
                            assert object_taxonomy.has_ability(s, "toggleable"), f"In transition rule file {json_file}, rule {rule}, {val} is not a toggleable machine."
                            assert object_taxonomy.has_ability(s, "fillable"), f"In transition rule file {json_file}, rule {rule}, {val} is not a fillable machine."

                    if key in CONTAINER_KEY:
                        val = rule[key]
                        for s in val:
                            assert object_taxonomy.has_ability(s, "fillable"), f"In transition rule file {json_file}, rule {rule}, {val} is not a fillable container."

def sanity_check_transition_rules_washer(object_taxonomy):
    leaf_synsets = object_taxonomy.get_leaf_descendants("entity.n.01")
    with open(TRANSITION_RULE_FOLDER / "washer.json", "r") as f:
        transition_rule = json.load(f)
    for system_synset, conditions in transition_rule.items():
        assert system_synset in leaf_synsets, f"In washer transition rule, {system_synset} is not a leaf synset."
        if conditions is not None:
            for cleanser_synset in conditions:
                assert cleanser_synset in leaf_synsets, f"In washer transition rule, {cleanser_synset} is not a leaf synset."

def sanity_check():
    # Lazy import so that it can use the latest version of output_hierarchy_properties.json
    from bddl.object_taxonomy import ObjectTaxonomy
    object_taxonomy = ObjectTaxonomy()
    sanity_check_object_hierarchy(object_taxonomy)
    sanity_check_transition_rules(object_taxonomy)
    sanity_check_transition_rules_washer(object_taxonomy)

if __name__ == '__main__':
    sanity_check()



import re
from nltk.corpus import wordnet as wn
from typing import Tuple, List, Set
from bddl.activity import get_initial_conditions, get_goal_conditions, get_object_scope
from bddl.logic_base import UnaryAtomicFormula, BinaryAtomicFormula, Expression
from bddl.backend_abc import BDDLBackend
from bddl.parsing import parse_domain

# STATE METADATA
STATE_MATCHED = "success"
STATE_PLANNED = "warning"
STATE_UNMATCHED = "danger"
STATE_SUBSTANCE = "info"
STATE_ILLEGAL = "secondary"
STATE_NONE = "light"


# predicates that can only be used for substances
SUBSTANCE_PREDICATES = {"filled", "insource", "empty", "saturated", "contains", "covered"}

# predicates that indicate the need for a fillable volume
FILLABLE_PREDICATES = {"filled", "contains", "empty"}

ANNOTATION_REQUIRED_PROPERTIES = {
    "fillable",
    "toggleable",
    "fireSource",
    # "sliceable",
    "slicer",
    "particleRemover",
    "particleApplier",
    "particleSource",
    "particleSink",
}


def canonicalize(s):
    try:
        return wn.synset(s).name()
    except:
        return s


def wn_synset_exists(synset):
  try:
    wn.synset(synset)
    return True
  except:
    return False
  

*__, domain_predicates = parse_domain("omnigibson")
UNARIES = [predicate for predicate, inputs in domain_predicates.items() if len(inputs) == 1]
BINARIES = [predicate for predicate, inputs in domain_predicates.items() if len(inputs) == 2]

class TrivialUnaryFormula(UnaryAtomicFormula):
    def _evaluate():
        return True 
    def _sample():
        return True
    

class TrivialBinaryFormula(BinaryAtomicFormula):
    def _evaluate():
        return True 
    def _sample():
        return True


def gen_unary_token(predicate_name, generate_ground_options=True):
    return type(f"{predicate_name}StateUnaryPredicate", (TrivialUnaryFormula,), {"STATE_CLASS": "HowDoesItMatter", "STATE_NAME": predicate_name})


def gen_binary_token(predicate_name, generate_ground_options=True):
    return type(f"{predicate_name}StateBinaryPredicate", (TrivialBinaryFormula,), {"STATE_CLASS": "HowDoesItMatter", "STATE_NAME": predicate_name})


class TrivialBackend(BDDLBackend):
    def get_predicate_class(self, predicate_name):
        if predicate_name in UNARIES: 
            return gen_unary_token(predicate_name)
        elif predicate_name in BINARIES:
            return gen_binary_token(predicate_name)
        else: 
            raise KeyError(predicate_name)


class TrivialGenericObject(object): 
    def __init__(self, name):
        self.name = name


def get_initial_and_goal_conditions(conds) -> Tuple[List, List]:
    scope = get_object_scope(conds)
    # Pretend scope has been filled 
    for name in scope: 
        scope[name] = TrivialGenericObject(name)
    initial_conds = get_initial_conditions(conds, TrivialBackend(), scope, generate_ground_options=False)
    goal_conds = get_goal_conditions(conds, TrivialBackend(), scope, generate_ground_options=False)
    return initial_conds, goal_conds

def get_leaf_conditions(cond) -> List:
    if isinstance(cond, list):
        return [leaf_cond for child in cond for leaf_cond in get_leaf_conditions(child)]
    # elif isinstance(cond, (UnaryAtomicFormula, BinaryAtomicFormula)):   # This is too slow.
    if hasattr(cond, "input") or hasattr(cond, "input1"):
        if cond.children:
            raise ValueError(f"Found an atomic formula {cond} with children.")
        else:
            return [cond]
    elif isinstance(cond, Expression):
        if not cond.children:
            raise ValueError(f"Found empty expression {cond} in tree.")
        
        return [leaf_cond for child in cond.children for leaf_cond in get_leaf_conditions(child)]
    else:
        raise ValueError(f"Found unexpected item {cond} in tree.")
    
def get_synsets(cond):
    def get_synset_from_scope_name(scope_name):
        lemma, n, number = scope_name.split(".")
        number = number.rsplit('_', 1)[0]
        synset = f"{lemma}.{n}.{number}"
        assert re.fullmatch(r'^[A-Za-z-_]+\.n\.[0-9]+$', synset), f"Invalid synset name: {synset}"
        return synset
    # TODO: Too slow!
    # assert isinstance(cond, (UnaryAtomicFormula, BinaryAtomicFormula)), "This only works with atomic formulae"
    if hasattr(cond, "input"):
        return [get_synset_from_scope_name(cond.input)]
    else:
        return [get_synset_from_scope_name(cond.input1), get_synset_from_scope_name(cond.input2)]


def object_substance_match(cond, synset) -> Tuple[bool, bool]:
    """
    Return two bools corresponding to whether synset is used as a non-substance and as a substance, respectively, in this condition subtree
    """
    leafs = get_leaf_conditions(cond)

    # It's used as a substance if it shows up as the last argument of any substance predicate
    is_used_as_substance = any(synset == get_synsets(leaf)[-1] for leaf in leafs if leaf.STATE_NAME in SUBSTANCE_PREDICATES)

    # It's used as a non-substance if it shows up as any argument of a non-substance predicate
    is_used_as_non_substance_in_non_substance_predicate = any(
        synset in get_synsets(leaf)
        for leaf in leafs
        if leaf.STATE_NAME not in SUBSTANCE_PREDICATES | {"future", "real"})
    # or the first argument of a two-argument substance predicate
    is_used_as_non_substance_in_substance_predicate = any(
        synset == get_synsets(leaf)[0]
        for leaf in leafs
        if leaf.STATE_NAME in SUBSTANCE_PREDICATES and hasattr(leaf, "input2"))
    is_used_as_non_substance = is_used_as_non_substance_in_non_substance_predicate or is_used_as_non_substance_in_substance_predicate
    return is_used_as_non_substance, is_used_as_substance



def object_used_as_fillable(cond, synset) -> Tuple[bool, bool]:
    """
    Return a bool corresponding to whether the synset is used as a fillable at any point
    """
    
    # Looking for the first argument of one of the fillable predicates.
    leafs = get_leaf_conditions(cond)
    return any(synset == get_synsets(leaf)[0] for leaf in leafs if leaf.STATE_NAME in FILLABLE_PREDICATES)
    

def object_used_predicates(cond, synset) -> Tuple[bool, bool]:
    leafs = get_leaf_conditions(cond)
    return {leaf.STATE_NAME for leaf in leafs if synset in get_synsets(leaf)}


def all_task_predicates(cond) -> Set[str]:
    return {leaf.STATE_NAME for leaf in get_leaf_conditions(cond)}


def leaf_inroom_conds(raw_cond, synsets: Set[str]) -> List[Tuple[str, str]]:
    """
    Return a list of all inroom conditions in the subtree of raw_cond
    """
    ret = []
    if isinstance(raw_cond, list):
        for child in raw_cond:
            ret.extend(leaf_inroom_conds(child, synsets))
        if raw_cond[0] == "inroom":
            synset = raw_cond[1].split('?')[-1].rsplit("_", 1)[0]
            assert synset in synsets, f"{synset} not in valid format"
            ret.append((canonicalize(synset), raw_cond[2]))
    return ret    



TM_SUBMAPS_TO_PARAMS = {
    "heat_cook": {
        "rule_name": {
            "required": False,
            "type": "string"
        },
        "input_synsets": {
            "required": True,
            "type": "synset"
        },
        "input_states": {
            "required": False,
            "type": "atom"
        },
        "container": {
            "required": False,
            "type": "synset"
        },
        # "container_input_relation": {
        #     "required"
        # },
        "heat_source": {
            "required": False,
            "type": "synset"
        },
        "output_synsets": {
            "required": True,
            "type": "synset"
        },
        "output_states": {
            "required": False,
            "type": "atom"
        },
        # "container_output_relation": {
        #     "required"
        # },
        "timesteps": {
            "required": False,
            "type": "integer"
        }
    },
    "mixing_stick": {
        "rule_name": {
            "required": False,
            "type": "string"
        },
        "input_synsets": {
            "required": True,
            "type": "synset"
        },
        "input_states": {
            "required": False,
            "type": "atom"
        },
        "output_synsets": {
            "required": True,
            "type": "synset"
        },
        "output_states": {
            "required": False,
            "type": "atom"

        }
    },
    "single_toggleable_machine": {
        "rule_name": {
            "required": False,
            "type": "string"
        },
        "input_synsets": {
            "required": True,
            "type": "synset"
        },
        "input_states": {
            "required": False,
            "type": "atom"
        },
        "machine": {
            "required": True,
            "type": "synset"

        },
        "output_synsets": {
            "required": True,
            "type": "synset"
        },
        "output_states": {
            "required": False,
            "type": "atom"
        }
    }
}




import pathlib
from bddl.data_generation.get_hierarchy_full import get_hierarchy, create_get_save_hierarchy_with_properties
from bddl.data_generation.get_syn_prop_annots_canonical import create_get_save_annots_canonical, create_get_save_properties_to_synsets, create_get_save_synsets_to_descriptors
from bddl.data_generation.propagate_by_intersection import create_get_save_propagated_canonical
from bddl.data_generation.process_prop_param_annots import create_get_save_propagated_annots_params
from bddl.data_generation.get_implicit_transition_rules import create_get_save_implicit_transition_rules
from bddl.data_generation.get_explicit_transition_rules import create_save_explicit_transition_rules
from bddl.data_generation.sanity_check import sanity_check
import pandas as pd
import csv
import nltk

'''
Inputs:
    ../house_room_info/currently_assigned_activities.json
    synset_masterlist.tsv
    TODO complete
'''
SYN_PROP_DATA_FN = pathlib.Path(__file__).parents[1] / "generated_data" / "synsets.csv"
# # Get owned models 

def main():
    nltk.download("wordnet")

    # Get full hierarchy (it's created and saved on import)
    with open(SYN_PROP_DATA_FN, "r") as f:
        syn_prop_dict = {}
        for row in csv.DictReader(f):
            syn_prop_dict[row.pop("synset")] = row
    hierarchy = get_hierarchy(syn_prop_dict)
    syn_prop_df = pd.read_csv(SYN_PROP_DATA_FN)

    # Create, get, and save master canonical 
    annots_canonical = create_get_save_annots_canonical(syn_prop_dict)
    # Create, get, and save final canonical from hierarchy, master_canonical
    propagated_canonical = create_get_save_propagated_canonical(hierarchy, annots_canonical)

    # Create and save properties_to_synsets from propagated canonical 
    props_to_syns = create_get_save_properties_to_synsets(propagated_canonical)

    # Create and save synsets_to_descriptors from propagated canonical
    create_get_save_synsets_to_descriptors(propagated_canonical)

    # Add parameter info to syns-to-props
    syns_to_param_props = create_get_save_propagated_annots_params(propagated_canonical)

    # Add prop-param info to hierarchy 
    create_get_save_hierarchy_with_properties(hierarchy)

    # Create and save implicit transition jsons
    create_get_save_implicit_transition_rules(syns_to_param_props, props_to_syns)

    # Create and save explicit transition jsons
    create_save_explicit_transition_rules()

    # Sanity check object taxonomy and transition rules
    sanity_check()

    # # Create and save activity-specific hierarchies (no getting because that will get complicated)
    # create_save_activity_specific_hierarchies()

    # # Create and save single dict of all activity-specific hierarchies. Since the other script is run before it, the files will be updated
    # create_save_all_activity_hierarchy_dict()

if __name__ == '__main__':
    main()


from collections import Counter
import copy
import csv
import glob
import json
import os
import pandas as pd
import pathlib
import re 
import sys

import bddl
from bddl.activity import Conditions
from bddl.data_generation.tm_submap_params import TM_SUBMAPS_TO_PARAMS
from bddl.parsing import parse_problem, parse_domain
from bddl.trivial_backend import *


# Files

BDDL_DIR = pathlib.Path(bddl.__file__).parent

PROBLEM_FILE_DIR = BDDL_DIR / "activity_definitions"
PROPS_TO_SYNS_JSON = BDDL_DIR / "generated_data/properties_to_synsets.json"
SYNS_TO_PROPS_JSON = BDDL_DIR / "generated_data/propagated_annots_canonical.json"
SYNS_TO_PROPS_PARAMS_JSON = BDDL_DIR / "generated_data/propagated_annots_params.json"
CSVS_DIR = "tm_csvs"
TRANSITION_MAP_DIR = BDDL_DIR / "generated_data/transition_map/tm_jsons"

# Constants

*__, domain_predicates = parse_domain("omnigibson")
UNARIES = [predicate for predicate, inputs in domain_predicates.items() if len(inputs) == 1]
BINARIES = [predicate for predicate, inputs in domain_predicates.items() if len(inputs) == 2]

VALID_ATTACHMENTS = set([
    ("mixing_bowl.n.01", "electric_mixer.n.01"),
    ("cork.n.04", "wine_bottle.n.01"),
    ("menu.n.01", "wall.n.01"),
    ("broken__light_bulb.n.01", "table_lamp.n.01"),
    ("light_bulb.n.01", "table_lamp.n.01"),
    ("lens.n.01", "digital_camera.n.01"),
    ("screen.n.01", "wall.n.01"),
    ("antler.n.01", "wall.n.01"),
    ("skateboard_wheel.n.01", "skateboard.n.01"),
    ("blackberry.n.01", "scrub.n.01"),
    ("raspberry.n.02", "scrub.n.01"),
    ("dip.n.07", "candlestick.n.01"),
    ("sign.n.02", "wall.n.01"),
    ("wreath.n.01", "wall.n.01"),
    ("bow.n.08", "wall.n.01"),
    ("holly.n.03", "wall.n.01"),
    ("curtain_rod.n.01", "wall.n.01"),
    ("bicycle.n.01", "bicycle_rack.n.01"),
    ("bicycle_rack.n.01", "wall.n.01"),
    ("dartboard.n.01", "wall.n.01"),
    ("rug.n.01", "wall.n.01"),
    ("fairy_light.n.01", "wall.n.01"),
    ("lantern.n.01", "wall.n.01"),
    ("address.n.05", "wall.n.01"),
    ("hanger.n.02", "wardrobe.n.01"),
    ("flagpole.n.02", "wall.n.01"),
    ("picture_frame.n.01", "wall.n.01"),
    ("wind_chime.n.01", "pole.n.01"),
    ("pole.n.01", "wall.n.01"),
    ("hook.n.05", "trailer_truck.n.01"),
    ("fire_alarm.n.02", "wall.n.01"),
    ("poster.n.01", "wall.n.01"),
    ("painting.n.01", "wall.n.01"),
    ("hanger.n.02", "coatrack.n.01"),
    ("license_plate.n.01", "car.n.01"),
    ("gummed_label.n.01", "license_plate.n.01"),
    ("wallpaper.n.01", "wall.n.01"),
    ("mirror.n.01", "wall.n.01"),
    ("webcam.n.02", "desktop_computer.n.01"),
    ("kayak.n.01", "kayak_rack.n.01"),
    ("kayak_rack.n.01", "wall.n.01"),
    ("trout.n.01", "fishing_rod.n.01"),
    ("bicycle_rack.n.01", "recreational_vehicle.n.01"),
])

VALID_ROOMS = set()

OBJECT_INSTANCE_RE = r"[A-Za-z-_]+\.n\.[0-9]+_[0-9]+"
OBJECT_CAT_RE = r"[A-Za-z-_]+\.n\.[0-9]+$"
OBJECT_CAT_AND_INST_RE = r"[A-Za-z-_]+\.n\.[0-9]+"
SINGLE_CAT_QUANTS = ["forall", "exists", "forn"]
DOUBLE_CAT_QUANTS = ["forpairs", "fornpairs"]
ROOMS = set([
    "kitchen", 
    "dining_room",
    "living_room", 
    "utility_room", 
    "empty_room", 
    "bedroom", 
    "bathroom", 
    "garden", 
    "shared_office", 
    "corridor", 
    "classroom", 
    "grocery_store",
    "computer_lab",
    "playroom",
    "sauna",
    "childs_room",
    "garage",
    "closet",
    "storage_room",
    "entryway",
    "private_office",
    "meeting_room",
    "bar",
    "staircase",
    "spa",
    "television_room",
    "lobby"
])
PLACEMENTS = set([
    # "connected",
    "ontop", 
    "inside", 
    "under", 
    "filled", 
    "covered", 
    "overlaid", 
    "saturated", 
    "inroom", 
    "insource", 
    # "hung", 
    "future",
    "attached",
    "draped",
    "contains"
])
SUBSTANCE_PLACEMENTS = set(["saturated", "filled", "covered", "insource", "contains"])
FUTURE_PREDICATE = "future"


# Helpers

def _traverse_goal_for_objects(expr, objects=None):
    objects = objects if objects is not None else set()
    # Check that category declarations in quantifiers are really categories, and equal
    if expr[0] in ["forall", "exists", "forpairs"]:
        term, __, cat = expr[1]
        assert term.strip("?") == cat, f"mismatched term and cat declaration: {term}, {cat}"
        assert re.match(OBJECT_CAT_RE, term.strip("?")) is not None, f"non-category term in quantifier declaration: {term}"
        if expr[0] in ["forpairs"]: 
            term, __, cat = expr[2]
            assert term.strip("?") == cat, f"mismatched term and cat declaration: {term}, {cat}"
            assert re.match(OBJECT_CAT_RE, term.strip("?")) is not None, f"non-category term in quantifier declaration: {term}"
        _traverse_goal_for_objects(expr[-1], objects=objects)
    if expr[0] in ["forn", "fornpairs"]:
        term, __, cat = expr[2]
        assert term.strip("?") == cat, f"mismatched term and cat declaration: {term}, {cat}"
        assert re.match(OBJECT_CAT_RE, term.strip("?")) is not None, f"non-category term in quantifier declaration: {term}"
        if expr[0] == "fornpairs": 
            term, __, cat = expr[3]
            assert term.strip("?") == cat, f"mismatched term and cat declaration: {term}, {cat}"
            assert re.match(OBJECT_CAT_RE, term.strip("?")) is not None, f"non-category term in quantifier declaration: {term}"
        _traverse_goal_for_objects(expr[-1], objects=objects)
    
    # Check the subexpr for atomic formulae in base case, else recurse 
    if type(expr[-1]) is not list: 
        for obj in expr[1:]:
            assert re.match(OBJECT_CAT_AND_INST_RE, obj.strip("?")) is not None, f"malformed object term in goal: {obj}"
            objects.add(obj.strip("?"))
    else: 
        if expr[0] in ["and", "or"]:
            for subexpr in expr[1:]:
                _traverse_goal_for_objects(subexpr, objects=objects)
        else:
            _traverse_goal_for_objects(expr[-1], objects=objects)


def _get_defn_elements_from_file(activity):
    defn_fn = os.path.join(PROBLEM_FILE_DIR, activity, 'problem0.bddl')
    with open(defn_fn, "r") as f:
        __, objects, init, goal = parse_problem(activity, 0, "omnigibson", predefined_problem=f.read())
    return activity, objects, init, goal


def _get_objects_from_object_list(objects):
    instances, categories = set(), set()
    for cat, insts in objects.items():
        categories.add(cat)
        for inst in insts: 
            instances.add(inst)
    return instances, categories


def _get_instances_in_init(init):
    '''
    Take a parsed :init condition and return a set of all instances in it.
    '''
    init_insts = set()
    for literal in init: 
        formula = literal[1] if literal[0] == "not" else literal
        for inst in formula[1:]: 
            assert (re.match(OBJECT_INSTANCE_RE, inst) is not None) or (inst in ROOMS), f":init has category: {inst}" 
            if inst not in ROOMS:
                init_insts.add(inst)
    return init_insts


def _get_objects_in_goal(goal):
    goal_objects = set()
    goal = ["and"] + goal
    _traverse_goal_for_objects(goal, goal_objects)
    return goal_objects


def _traverse_goal_for_atoms(expr, goal_atoms):
    if all(type(subexpr) == str for subexpr in expr):
        goal_atoms.append(expr)
    elif expr[0] in ["and", "or"]:
        for subexpr in expr[1:]:
            _traverse_goal_for_atoms(subexpr, goal_atoms)
    elif expr[0] in ["forall", "exists", "forn", "forpairs", "fornpairs"]:
        _traverse_goal_for_atoms(expr[-1], goal_atoms)
    elif expr[0] == "imply":
        _traverse_goal_for_atoms(expr[1], goal_atoms)
        _traverse_goal_for_atoms(expr[2], goal_atoms)
    elif expr[0] == "not":
        _traverse_goal_for_atoms(expr[1], goal_atoms)
    else:
        raise ValueError(f"Unhandled logic operator {expr[0]}")


def _get_atoms_in_goal(goal):
    goal_atoms = []
    for goal_expr in goal:
        _traverse_goal_for_atoms(goal_expr, goal_atoms)
    return goal_atoms


def _get_unique_items_from_transition_map():
    obj_set = set()
    for fname in glob.glob(CSVS_DIR):
        with open(fname) as f:
            for row in f:
                first = row.split(',')[0]
                if '.n.' in first:
                    obj_set.add(first.rpartition('_')[0])

    obj_set.remove('')
    
    for obj in obj_set:
        print(obj)


def is_specific_container_synset(synset): 
    return "__" in synset and "__of__" not in synset and "diced__" not in synset and "cooked__" not in synset and "half__" not in synset


def check_synset_predicate_alignment(atom, syns_to_props):
    if atom[0] == "ontop" and atom[1] == "agent.n.01_1":
        return 

    pred, *object_insts = atom 
    objects = []
    for object_inst in object_insts: 
        syn_match = re.match(OBJECT_CAT_AND_INST_RE, object_inst.strip("?"))
        if syn_match is not None:
            objects.append(syn_match.group())
        elif True: # object_inst in VALID_ROOMS:    # TODO uncomment when VALID_ROOMS is populated
            if pred == "inroom":
                objects.append(object_inst)
            else:
                raise AssertionError(f"Nonsynset {object_inst} outside inroom")
        else:
            raise AssertionError(f"Invalid room {object_inst}")
    assert (pred in UNARIES) or (pred in BINARIES), f"Invalid predicate: {pred}"
    assert ((pred in UNARIES) and (len(objects) == 1)) or ((pred in BINARIES) and (len(objects) == 2)), f"Atom has wrong arity: {atom}"

    if pred == "cooked":
        assert "nonSubstance" in syns_to_props[objects[0]], f"Inapplicable cooked: {atom}"
        assert "cookable" in syns_to_props[objects[0]], f"Inapplicable cooked: {atom}"
    if pred == "frozen":
        assert "nonSubstance" in syns_to_props[objects[0]], f"Inapplicable frozen: {atom}"
        assert "freezable" in syns_to_props[objects[0]], f"Inapplicable frozen: {atom}"
    if pred == "closed" or pred == "open":
        assert "rigidBody" in syns_to_props[objects[0]], f"Inapplicable closed/open: {atom}"
        assert "openable" in syns_to_props[objects[0]], f"Inapplicable closed/open: {atom}"
    if pred == "folded" or pred == "unfolded":
        # cloth or rope is drapeable
        assert "drapeable" in syns_to_props[objects[0]], f"Inapplicable folded/unfolded: {atom}"
    if pred == "toggled_on":
        assert "rigidBody" in syns_to_props[objects[0]], f"Inapplicable toggled_on: {atom}"
        assert "toggleable" in syns_to_props[objects[0]], f"Inapplicable toggled_on: {atom}"
    if pred == "hot":
        assert "nonSubstance" in syns_to_props[objects[0]], f"Inapplicable hot: {atom}"
        assert "heatable" in syns_to_props[objects[0]], f"Inapplicable hot: {atom}"
    if pred == "on_fire":
        assert "nonSubstance" in syns_to_props[objects[0]], f"Inapplicable on_fire: {atom}"
        assert "flammable" in syns_to_props[objects[0]], f"Inapplicable on_fire: {atom}"
    if pred == "assembled":
        assert "rigidBody" in syns_to_props[objects[0]], f"Inapplicable assembled: {atom}"
        assert "assembleable" in syns_to_props[objects[0]], f"Inapplicable assembled: {atom}"
    if pred == "broken":
        assert "rigidBody" in syns_to_props[objects[0]], f"Inapplicable broken: {atom}"
        assert "breakable" in syns_to_props[objects[0]], f"Inapplicable broken: {atom}"

    # Binaries
    if pred == "saturated":
        assert ("nonSubstance" in syns_to_props[objects[0]]) and ("particleRemover" in syns_to_props[objects[0]]) and ("substance" in syns_to_props[objects[1]]), f"Inapplicable saturated: {atom}"
    if pred == "covered":
        assert ("nonSubstance" in syns_to_props[objects[0]]) and ("substance" in syns_to_props[objects[1]]), f"Inapplicable covered: {atom}"
        if "drapeable" in syns_to_props[objects[0]]:
            assert "visualSubstance" in syns_to_props[objects[1]], f"Inapplicable covered: {atom}"
    if pred == "filled":
        assert ("rigidBody" in syns_to_props[objects[0]]) and ("fillable" in syns_to_props[objects[0]]) and ("physicalSubstance" in syns_to_props[objects[1]]), f"Inapplicable filled/empty: {atom}"
    if pred == "contains" or pred == "empty":
        assert ("rigidBody" in syns_to_props[objects[0]]) and ("fillable" in syns_to_props[objects[0]]) and ("substance" in syns_to_props[objects[1]]), f"Inapplicable contains: {atom}"
    if pred == "ontop":
        assert ("nonSubstance" in syns_to_props[objects[0]]) and ("rigidBody" in syns_to_props[objects[1]] or "softBody" in syns_to_props[objects[1]]), f"Inapplicable ontop: {atom}"
    if pred == "nextto":
        assert ("nonSubstance" in syns_to_props[objects[0]]) and ("nonSubstance" in syns_to_props[objects[1]]), f"Inapplicable nextto: {atom}"
    if pred == "under":
        assert ("nonSubstance" in syns_to_props[objects[0]]) and ("rigidBody" in syns_to_props[objects[1]] or "softBody" in syns_to_props[objects[1]]), f"Inapplicable under: {atom}"
    if pred == "touching": 
        assert ("rigidBody" in syns_to_props[objects[0]]) and ("rigidBody" in syns_to_props[objects[1]]), f"Inapplicable touching: {atom}"
    if pred == "inside": 
        assert ("nonSubstance" in syns_to_props[objects[0]]) and ("rigidBody" in syns_to_props[objects[1]] or "softBody" in syns_to_props[objects[1]]), f"Inapplicable inside: {atom}"
    if pred == "overlaid": 
        assert ("drapeable" in syns_to_props[objects[0]]) and ("rigidBody" in syns_to_props[objects[1]]), f"Inapplicable overlaid: {atom}"
    if pred == "attached":
        assert tuple(objects) in VALID_ATTACHMENTS, f"Inapplicable attached: {atom}"
    if pred == "draped": 
        assert ("drapeable" in syns_to_props[objects[0]]) and ("rigidBody" in syns_to_props[objects[1]]), f"Inapplicable overlaid: {atom}"
    if pred == "insource": 
        assert (("particleSource" in syns_to_props[objects[0]]) or ("particleApplier" in syns_to_props[objects[0]])) and ("substance" in syns_to_props[objects[1]]), f"Inapplicable insource: {atom}"


def check_clashing_transition_rules():
    # Check within each submap
    for submap_name in TM_SUBMAPS_TO_PARAMS:
        with open(os.path.join("..", "bddl", "generated_data", "transition_map", "tm_jsons", submap_name + ".json"), "r") as f:
            submap = json.load(f)
        seen_object_sets = []
        for rule in submap: 
            # Get relevant parameters
            rule_name = rule.get("rule_name", "No name")
            input_objects = rule.get("input_synsets", {})
            input_states = rule.get("input_states", {})
            input_states = input_states if input_states is not None else {}
            if submap_name == "heat_cook": 
                equipment = set([list(rule["heat_source"].keys())[0], list(rule["container"].keys())[0]])
            elif submap_name == "single_toggleable_machine":
                equipment = set([list(rule["machine"].keys())[0]])
            else:
                equipment = set()       # Equivalence will pass trivially when checked, because this rule already clashes
            output_objects = rule.get("output_synsets", {})
            output_states = rule.get("output_states", {})
            output_states = output_states if output_states is not None else {}

            # NOTE doing input_objects.keys, not input_objects.items, because simulator is not actually sensitive to amount. It only checks for category, 
            #   so different amounts but same categories still need to result in the same output.
            
            # Collect all sets of input objects to check subsetting
            input_objects = set(sorted(input_objects.keys(), key=lambda x: x[0]))
            output_objects = set(sorted(output_objects.keys(), key=lambda x: x[0]))
            input_states_strs = set([syns + "@" + ";".join([f"{pred}:{val}" for pred, val in sorted(states, key=lambda x: x[0])]) for syns, states in sorted(input_states.items(), key=lambda x: x[0])])
            output_states_strs = set([syns + "@" + ";".join([f"{pred}:{val}" for pred, val in sorted(states, key=lambda x: x[0])]) for syns, states in sorted(output_states.items(), key=lambda x: x[0])])
            for seen_rule_name, seen_object_set, seen_states_set, seen_equipment, seen_output_objects, seen_output_states in seen_object_sets:
                # If we see that our next input objects set is a subset or superset...
                if input_objects.issuperset(seen_object_set) or input_objects.issubset(seen_object_set):
                    # Construct a set of atomic formulae in string format
                    if input_states_strs.issuperset(seen_states_set) or input_states_strs.issubset(seen_states_set):
                        if equipment == seen_equipment:
                            # At this point, the output needs to be identical
                            if not output_objects == seen_output_objects or not output_states_strs == seen_output_states:
                                raise AssertionError(f"Clashing rules with input objects {rule_name} and {seen_rule_name} in submap {submap_name}.")
                
            seen_object_sets.append((rule_name, input_objects, input_states_strs, equipment, output_objects, output_states_strs))

    # Check heat_cook rules vs cooking individual items, since those are a potential clash that we know of
    # for activity in os.listdir(PROBLEM_FILE_DIR):
    #     if not os.path.isdir(os.path.join(PROBLEM_FILE_DIR, activity)): continue
    #     __, objects, init, goal = _get_defn_elements_from_file(activity)
        
    #     # Check for a rigid body ending cooked and starting not cooked/cook state unspecified, or a future clause for a cooked substance
    #     for cooking_rigids: 


    # NOTE other heuristics as we discover them


# BDDL tests 

def is_water_sourced(objects, init, goal):
    obj_list = []
    for obj_cat in objects:
        obj_list.extend(objects[obj_cat])
    objs = set(obj_list)
    # NOTE Assumes that if water is in goal, it'll be in objects
    if "water.n.06_1" in objs:
        # Find a placement for the water in the init
        for atom in init: 
            # match = re.match(r'^\((filled|covered) water\.n\.06_1', atom) is not None
            placed = (atom[0] in set(["filled", "covered"])) and (atom[1] == "water.n.06_1")
            # match_insource = re.match(r'^\(insource water\.n\.06_1 sink\.n\.01_[0-9]+\)$', atom) is not None
            insourced = (atom[0] == "insource") and (atom[1] == "water.n.06_1") and (re.match(r"sink\.n\.01_[0-9]+", atom[2]) is not None)
            if placed:
                # Make sure the container is in the object list 
                container = atom[-1]
                if container in objs: 
                    print("water was in container")
                    break 
                else:
                    # print(container)
                    # print(objs)
                    raise ValueError("Water is filling/covering something in the init, but that thing is not in the objects section.")
            if insourced:
                # Make sure the sink is in the object list
                sink = atom[-1]
                if sink not in objs:
                    raise ValueError("Water is insource a sink, but that sink is not in the objects section.")
                # Make sure the sink is in a room
                for atom in init: 
                    if (atom[0] == "inroom" and atom[1] == sink):
                        print("water was in sink and sink was in room")
                        return True
                else:
                    raise ValueError("Water is insource a sink and sink is in objects section, but not in an inroom")
    else:
        print("no water")
    return True


# Checkers

def object_list_correctly_formatted(defn):
    '''
    Verify that object list has the following properties for each line: 
    - Each ending term matches category regex
    - Penultimate term is `-`
    - First through third-to-last terms match instance regex
    - First through third-to-last terms are of the same category as the ending term
    '''
    objects_section = defn.split("(:objects\n")[-1].split("\n    )")[0].split("\n")
    for line in objects_section:
        elements = line.strip().split(" ")
        category = elements[-1]
        assert re.match(OBJECT_CAT_RE, category) is not None, f"Malformed category at end of object section line: {category}"
        assert elements[-2] == "-", f"There should be a hyphen but instead there is {elements[-2]}"
        for inst in elements[:-2]:
            assert re.match(OBJECT_INSTANCE_RE, inst) is not None, f"Malformed instance {inst}"
            assert category == re.match(OBJECT_CAT_AND_INST_RE, inst).group(), f"Mismatched category and object: {category} and {inst}"


def all_objects_appropriate(objects, init, goal):
    '''
    Checks the following: 
    1. All objects in :objects are in :init
    2. All objects in :init are in :objects (so basically, the set of objects in :objects and 
        the set of objects in :init are equivalent)
    3. The set of object instances and categories in :goal is a subset of the set of object 
        instances and categories in :objects
    4. There are no categories in :init
    '''
    instances, categories = _get_objects_from_object_list(objects)
    init_insts = _get_instances_in_init(init)
    
    assert init_insts.issubset(instances), f":init has object instances not in :objects: {init_insts.difference(instances)}"
    assert instances.issubset(init_insts), f":objects has object instances not in :init: {instances.difference(init_insts)}"
    
    goal_objects = _get_objects_in_goal(goal)
    assert goal_objects.issubset(categories.union(instances)), f":goal has objects not in :objects: {goal_objects.difference(categories.union(instances))}"


def all_objects_placed(init): 
    '''
    Make sure everything is placed relative to a ROOM, even transitively.
    This is because when I hand-edit, I may accidentally not put every new scene object in an inroom.
    Meanwhile, it's unlikely that TODO fill in whatever else was I thinking of that this function won't check
        but is unlikely lol.
        - Appropriate placements for the object type - I'm unlikely to mess that up if I do say so myself

    NOTE should only be executed AFTER all_objects_appropraite
    '''
    insts = _get_instances_in_init(init)
    insts = set([inst for inst in insts if ["future", inst] not in init])

    # Make sure everything not set to `future` is placed relative to a ROOM
    placed_insts = set()
    old_placed_insts = set()
    saturated = False 
    while not saturated:
        for inst in insts:                 
            if inst in placed_insts:
                continue
            for literal in init: 
                formula = literal[1] if literal[0] == "not" else literal 
                # NOTE only uncomment below line suffix when dealing with situations where substance and object have been flipped
                if (formula[0] == FUTURE_PREDICATE and formula[1] == inst) or ((formula[0] in PLACEMENTS) and (formula[1] == inst) and ((formula[2] in ROOMS) or (formula[2] in placed_insts))) or ((formula[0] in SUBSTANCE_PLACEMENTS) and (formula[1] in placed_insts) and (formula[2] == inst)):
                    placed_insts.add(inst)
        saturated = old_placed_insts == placed_insts 
        old_placed_insts = copy.deepcopy(placed_insts)
    assert not placed_insts.difference(insts), "There are somehow placed insts that are not in the overall set of insts."
    assert placed_insts == insts, f"Unplaced object instances: {insts.difference(placed_insts)}"


def no_invalid_synsets(objects, init, goal, syns_to_props):
    instances, categories = _get_objects_from_object_list(objects)
    init_insts = _get_instances_in_init(init)
    goal_objects = _get_objects_in_goal(goal)
    object_insts = set([re.match(OBJECT_CAT_AND_INST_RE, inst).group() for inst in instances.union(init_insts).union(goal_objects)])
    object_terms = object_insts.union(categories)
    for proposed_syn in object_terms: 
        assert (proposed_syn in syns_to_props) or (proposed_syn == "agent.n.01"), f"Invalid synset: {proposed_syn}"


def no_invalid_predicates(init, goal, domain_predicates):
    atoms = []
    for literal in init: 
        atoms.append(literal[1] if literal[0] == "not" else literal)
    atoms.extend(_get_atoms_in_goal(goal))
    for atom in atoms: 
        assert atom[0] in domain_predicates, f"Invalid predicate: {atom[0]}" 
        

# Check uncontrolled categories

def future_and_real_present(objects, init, goal): 
    init_objects = _get_instances_in_init(init)
    future_objects = set([literal[1] for literal in init if literal[0] == "future"])
    real_objects = set([expression[1].strip("?") for expression in goal if expression[0] == "real"])
    for expression in goal:
        if expression[0] == "or":
            for disjunct in expression[1:]:
                if disjunct[0] == "real":
                    real_objects.add(disjunct[1].strip("?"))
    assert future_objects.issubset(real_objects) or (("washer.n.03" in objects) and future_objects.difference(real_objects) == set(["water.n.06_1"])), f"{future_objects.difference(real_objects)} in future clauses but not real clauses (and doesn't satisfy washer/water exception)"
    assert real_objects.issubset(future_objects.union(init_objects)), f"{real_objects.difference(future_objects)} in real clauses but not future clauses or init"


def no_repeat_object_lines(defn): 
    from pprint import pprint
    objects_lines = [line.strip() for line in defn.split("(:objects")[1].split("(:init")[0].split("\n")[1:-3]]
    seen_cats = set() 
    for objects_line in objects_lines:
        if objects_line == ")":
            continue
        *__, __, cat = objects_line.split(" ")
        assert cat not in seen_cats, f"Category repeated: {cat}"
        seen_cats.add(cat)


def no_qmarks_in_init(defn): 
    init = defn.split("(:init")[1].split("(:goal")[0]
    assert "?" not in init, "Inappropriate ? in :init."


def no_contradictory_init_atoms(init):
    for literal in init: 
        if literal[0] == "not":
            assert literal[1] not in init, f"Contradictory init statements: {literal[1]}"


def no_uncontrolled_category(activity, defn):
    conds = bddl.activity.Conditions(activity, 0, "omnigibson", predefined_problem=defn)
    scope = bddl.activity.get_object_scope(conds)
    bddl.activity.get_initial_conditions(conds, TrivialBackend(), scope, generate_ground_options=False)
    # Pretend scope has been filled 
    for name in scope: 
        scope[name] = TrivialGenericObject(name, TrivialSimulator())
    bddl.activity.get_goal_conditions(conds, TrivialBackend(), scope, generate_ground_options=False)


def agent_present(init):
    for literal in init: 
        if (literal[0] == "ontop") and (literal[1] == "agent.n.01_1"):
            break
    else:
        raise AssertionError("Agent not present.")


def problem_name_correct(activity, definition_id=0):
    defn_fn = os.path.join(PROBLEM_FILE_DIR, activity, f'problem{definition_id}.bddl')
    with open(defn_fn, "r") as f:
        problem_name, *__ = parse_problem(activity, 0, "omnigibson", predefined_problem=f.read())
    assert (problem_name == f"{activity}-{definition_id}") or (problem_name == f"{activity.lower()}-{definition_id}"), f"Wrong problem name '{problem_name}' for activity '{activity}'"


def no_misaligned_synsets_predicates(init, goal, syns_to_props):
    for literal in init: 
        init_atom = literal[1] if literal[0] == "not" else literal
        check_synset_predicate_alignment(init_atom, syns_to_props)
    goal_atoms = _get_atoms_in_goal(goal)
    for goal_atom in goal_atoms:
        check_synset_predicate_alignment(goal_atom, syns_to_props)


def no_unnecessary_specific_containers(objects, init, goal, syns_to_props):
    specific_fillable_containers = [obj_cat for obj_cat in objects.keys() if obj_cat != "agent.n.01" and "fillable" in syns_to_props[obj_cat] and is_specific_container_synset(obj_cat)]
    
    atoms = []
    for literal in init: 
        atoms.append(literal[1] if literal[0] == "not" else literal)
    goal_atoms = [[term.strip("?") for term in goal_atom] for goal_atom in _get_atoms_in_goal(goal)]
    atoms.extend(goal_atoms)
    fill_atoms = [atom for atom in atoms if (atom[0] in ["filled", "contains", "insource", "inside"])]

    for specific_fillable_container in specific_fillable_containers:
        for atom in fill_atoms: 
            # print(atom)
            if (atom[0] in ["filled", "contains", "insource"]) and (re.match(OBJECT_CAT_AND_INST_RE, atom[1]).group() == specific_fillable_container):
                break 
            if (atom[0] == "inside") and (re.match(OBJECT_CAT_AND_INST_RE, atom[2]).group() == specific_fillable_container):
                break
        else:
            raise AssertionError(f"Substance-specific fillable container {specific_fillable_container} that does not fill/contain anything/have anything inside. Switch to container__of version.")


def no_substances_with_multiple_instances(objects, syns_to_props):
    for cat, insts in objects.items(): 
        if "substance" in syns_to_props[cat]:
            assert len(insts) == 1, f"Substance {cat} with {len(insts)} instances instead of 1"


# Transition map tests

def no_missing_required_params(rule, submap):
    param_metadata = TM_SUBMAPS_TO_PARAMS[submap]
    for param in param_metadata: 
        assert (not param_metadata[param]["required"]) or (rule[param] is not None), f"Required param {param} of rule {rule['rule_name']} is required but not present."
    for param, value in rule.items(): 
        if param_metadata[param]["required"]:
            assert value is not None, f"Required param {param} of rule {rule['rule_name']} is required has no value"


def no_incorrectly_formatted_params(rule, submap):
    param_metadata = TM_SUBMAPS_TO_PARAMS[submap]
    for param, value in rule.items(): 
        if param_metadata[param]["type"] == "synset" and value is not None:
            assert type(value) == dict, f"Malformed synset-type value for param {param} in rule {rule['rule_name']}"
            for proposed_synset, proposed_integer in value.items():
                assert re.match(OBJECT_CAT_RE, proposed_synset) is not None, f"Malformed synset {proposed_synset} in param {param} of rule {rule['rule_name']} in submap {submap}"
                assert type(proposed_integer) == int, f"Malformed integer {proposed_integer} in param {param} of rule {rule['rule_name']}"
        elif param_metadata[param]["type"] == "atom" and value is not None:
            assert type(value) == dict, f"Malformed atom-type value for param {param} in rule {rule['rule_name']} in submap {submap}"
            for proposed_synsets, proposed_predicates_values in value.items():
                for proposed_synset in proposed_synsets.split(","):
                    assert re.match(OBJECT_CAT_RE, proposed_synset) is not None, f"Malformed synset {proposed_synset} in param {param} of rule {rule['rule_name']} in submap {submap}"
                for proposed_predicate_value in proposed_predicates_values:
                    assert len(proposed_predicate_value) == 2, f"Malformed predicate-value pair {proposed_predicate_value} for param {param} in rule {rule['rule_name']}"
                    predicate, val = proposed_predicate_value 
                    assert type(predicate) == str, f"Malformed predicate {predicate} for param {param} in rule {rule['rule_name']} in submap {submap}"
                    assert type(val) == bool, f"Malformed predicate value {val} for param {param} in rule {rule['rule_name']} in submap {submap}"
        elif param_metadata[param]["type"] == "integer": continue
        elif param_metadata[param]["type"] == "string": continue
        elif value is None: 
            continue
        else:
            raise ValueError(f"Unhandled parameter type {param_metadata[param]['type']}")
        

def no_invalid_synsets_tm(rule, submap, syns_to_props):
    param_metadata = TM_SUBMAPS_TO_PARAMS[submap]
    for param, value in rule.items(): 
        if param_metadata[param]["type"] == "synset" and value is not None:   # we know the format since we enforced it above
            for proposed_synset in value.keys():
                assert proposed_synset in syns_to_props, f"Invalid synset: {proposed_synset} in rule {rule['rule_name']} in submap {submap}"
        elif param_metadata[param]["type"] == "atom" and value is not None:
            for proposed_synsets in value.keys():
                for proposed_synset in proposed_synsets.split(","):
                    assert proposed_synset in syns_to_props, f"Invalid synset {proposed_synset} in rule {rule['rule_name']} in submap {submap}"
        elif param_metadata[param]["type"] == "integer": continue
        elif param_metadata[param]["type"] == "string": continue
        elif value is None:
            continue
        else:
            raise ValueError(f"Unhandled parameter type {param_metadata[param]['type']}")
        

def no_invalid_predicates_tm(rule, submap, domain_predicates):
    param_metadata = TM_SUBMAPS_TO_PARAMS[submap]
    for param, value in rule.items(): 
        if param_metadata[param]["type"] == "synset": continue 
        elif param_metadata[param]["type"] == "integer": continue
        elif param_metadata[param]["type"] == "string": continue
        elif value is None: continue
        elif param_metadata[param]["type"] == "atom": 
            for __, proposed_predicate_values in value.items(): 
                for proposed_predicate, __ in proposed_predicate_values:
                    assert proposed_predicate in domain_predicates, f"Invalid predicate {proposed_predicate} in rule {rule['rule_name']} in submap {submap}"
        else:
            raise ValueError(f"Unhandled parameter type {param_metadata[param]['type']}")


def no_misaligned_synsets_predicates_tm(rule, submap, syns_to_props):
    param_metadata = TM_SUBMAPS_TO_PARAMS[submap]
    for param, value in rule.items(): 
        if param_metadata[param]["type"] == "synset": continue 
        elif param_metadata[param]["type"] == "integer": continue
        elif param_metadata[param]["type"] == "string": continue
        elif value is None: continue
        elif param_metadata[param]["type"] == "atom": 
            for synsets, predicate_vals in value.items(): 
                for predicate, __ in predicate_vals:
                    synsets = synsets.split(",")
                    check_synset_predicate_alignment([predicate, *synsets], syns_to_props)
        else:
            raise ValueError(f"Unhandled parameter type {param_metadata[param]['type']}")


def no_substances_with_multiple_instances_tm(rule, submap, syns_to_props):
    param_metadata = TM_SUBMAPS_TO_PARAMS[submap]
    for param, value in rule.items(): 
        if param_metadata[param]["type"] == "synset":
            for synset, num_instances in value.items(): 
                if "substance" in syns_to_props[synset]:
                    assert num_instances == 1, f"Substance {synset} with {num_instances} instances instead of 1 in rule {rule['rule_name']} in submap {submap}"
        elif param_metadata[param]["type"] == "atom": continue
        elif param_metadata[param]["type"] == "integer": continue
        elif param_metadata[param]["type"] == "string": continue
        else:
            raise ValueError(f"Unhandled parameter type {param_metadata[param]['type']}")

def no_duplicate_rule_names():
    # Get the JSON files
    json_paths = glob.glob(os.path.join(TRANSITION_MAP_DIR, "*.json"))
    data = []
    for jp in json_paths:
        # Washer rule is a special case
        if "washer" in jp:
            continue
        with open(jp) as f:
            data.append(json.load(f))
    transitions = [rule for rules in data for rule in rules]
    rule_names = [rule["rule_name"] for rule in transitions]
    repeated_rule_names = [k for k, v in Counter(rule_names).items() if v > 1]
    if repeated_rule_names:
        raise ValueError(f"Repeated rule names: {repeated_rule_names}") 
    


from abc import abstractmethod, ABCMeta

from future.utils import with_metaclass
from bddl.utils import UncontrolledCategoryError


class Expression(with_metaclass(ABCMeta)):
    def __init__(self, scope, backend, body, object_map):
        self.children = []
        self.child_values = []
        self.kwargs = {}
        self.body = body
        self.scope = scope
        self.backend = backend
        self.object_map = object_map

    @abstractmethod
    def evaluate(self):
        pass


class AtomicFormula(Expression):
    def __init__(self, scope, backend, body, object_map):
        super().__init__(scope, backend, body, object_map)


class BinaryAtomicFormula(AtomicFormula):
    STATE_NAME = None

    def __init__(self, scope, backend, body, object_map, generate_ground_options=True):
        super().__init__(scope, backend, body, object_map)
        assert len(body) == 2, f'Param list for predicate {self.STATE_NAME} should have 2 args'
        self.input1, self.input2 = [inp.strip('?') for inp in body]
        self.scope = scope
        try:
            if isinstance(self.scope[self.input1], str):
                self.input1 = self.scope[self.input1]
        except KeyError as e:
            raise UncontrolledCategoryError(e)
        try:
            if isinstance(self.scope[self.input2], str):
                self.input2 = self.scope[self.input2]
        except KeyError as e:
            raise UncontrolledCategoryError(e)

        if generate_ground_options:
            self.get_ground_options()

    @abstractmethod
    def _evaluate(self, obj1, obj2):
        pass

    def evaluate(self):
        if (self.scope[self.input1] is not None) and (self.scope[self.input2] is not None):
            return self._evaluate(self.scope[self.input1], self.scope[self.input2], **self.kwargs)
        else:
            print('%s and/or %s are not mapped to simulator objects in scope' %
                  (self.input1, self.input2))

    @abstractmethod
    def _sample(self, obj1, obj2, binary_state):
        pass

    def sample(self, binary_state):
        if (self.scope[self.input1] is not None) and (self.scope[self.input2] is not None):
            return self._sample(self.scope[self.input1], self.scope[self.input2], binary_state, **self.kwargs)
        else:
            print('%s and/or %s are not mapped to simulator objects in scope' %
                  (self.input1, self.input2))

    def get_ground_options(self):
        self.flattened_condition_options = [
            [[self.STATE_NAME, self.input1, self.input2]]]


class UnaryAtomicFormula(AtomicFormula):
    STATE_NAME = None

    def __init__(self, scope, backend, body, object_map, generate_ground_options=True):
        super().__init__(scope, backend, body, object_map)
        assert len(body) == 1, f'Param list for predicate {self.STATE_NAME} should have 1 arg'
        self.input = body[0].strip('?')
        self.scope = scope
        try:
            if isinstance(self.scope[self.input], str):
                self.input = self.scope[self.input]
        except KeyError as e:
            raise UncontrolledCategoryError(e)
        
        if generate_ground_options:
            self.get_ground_options()

    @abstractmethod
    def _evaluate(self, obj):
        pass

    def evaluate(self):
        if self.scope[self.input] is not None:
            return self._evaluate(self.scope[self.input], **self.kwargs)
        else:
            print('%s is not mapped to a simulator object in scope' % self.input)
            return False

    @abstractmethod
    def _sample(self, obj, binary_state):
        pass

    def sample(self, binary_state):
        if self.scope[self.input] is not None:
            return self._sample(self.scope[self.input], binary_state, **self.kwargs)
        else:
            print('%s is not mapped to a simulator object in scope' % self.input)
            return False

    def get_ground_options(self):
        self.flattened_condition_options = [
            [[self.STATE_NAME, self.input]]]




import unittest

from bddl import object_taxonomy


class TaxonomyTest(unittest.TestCase):
    def setUp(self):
        self.taxonomy = object_taxonomy.ObjectTaxonomy()

    def test_get_class_name_from_igibson_category(self):
        self.assertIsNotNone(self.taxonomy.get_class_name_from_igibson_category("picture"))
        self.assertIsNone(self.taxonomy.get_class_name_from_igibson_category("invalid_category"))

    def test_get_subtree_igibson_categories(self):
        self.assertIn("apple", self.taxonomy.get_subtree_igibson_categories("fruit.n.01"))
        self.assertNotIn("potato", self.taxonomy.get_subtree_igibson_categories("fruit.n.01"))

    def test_is_valid_class(self):
        self.assertTrue(self.taxonomy.is_valid_class("entity.n.01"))
        self.assertFalse(self.taxonomy.is_valid_class("invalid_class"))

    def test_get_descendants(self):
        descendants = self.taxonomy.get_descendants("home_appliance.n.01")
        self.assertIn("kitchen_appliance.n.01", descendants)
        self.assertIn("stove.n.01", descendants)
        self.assertNotIn("entity.n.01", descendants)

    def test_get_leaf_descendants(self):
        descendants = self.taxonomy.get_leaf_descendants("home_appliance.n.01")
        self.assertNotIn("kitchen_appliance.n.01", descendants)
        self.assertNotIn("entity.n.01", descendants)
        self.assertIn("stove.n.01", descendants)

    def test_get_ancestors(self):
        ancestors = self.taxonomy.get_ancestors("kitchen_appliance.n.01")
        self.assertIn("home_appliance.n.01", ancestors)
        self.assertIn("entity.n.01", ancestors)
        self.assertNotIn("stove.n.01", ancestors)

    def test_is_descendant(self):
        self.assertTrue(self.taxonomy.is_descendant("kitchen_appliance.n.01", "home_appliance.n.01"))
        self.assertTrue(self.taxonomy.is_descendant("kitchen_appliance.n.01", "entity.n.01"))
        self.assertFalse(self.taxonomy.is_descendant("kitchen_appliance.n.01", "stove.n.01"))

    def test_is_ancestor(self):
        self.assertFalse(self.taxonomy.is_ancestor("kitchen_appliance.n.01", "home_appliance.n.01"))
        self.assertFalse(self.taxonomy.is_ancestor("kitchen_appliance.n.01", "entity.n.01"))
        self.assertTrue(self.taxonomy.is_ancestor("kitchen_appliance.n.01", "stove.n.01"))

    def test_get_abilities(self):
        self.assertIn("heatSource", self.taxonomy.get_abilities("stove.n.01"))
        self.assertNotIn("burnable", self.taxonomy.get_abilities("stove.n.01"))

    def test_get_igibson_categories(self):
        self.assertIn("standing_tv", self.taxonomy.get_igibson_categories("television_receiver.n.01"))
        self.assertIn("wall_mounted_tv", self.taxonomy.get_igibson_categories("television_receiver.n.01"))
        self.assertNotIn("stove", self.taxonomy.get_igibson_categories("television_receiver.n.01"))

    def test_get_children(self):
        self.assertIn("kitchen_appliance.n.01", self.taxonomy.get_children("home_appliance.n.01"))
        self.assertNotIn("stove.n.01", self.taxonomy.get_children("home_appliance.n.01"))
        self.assertFalse(len(self.taxonomy.get_children("stove.n.01")))

    def test_get_parent(self):
        self.assertEqual(self.taxonomy.get_parent("stove.n.01"), "kitchen_appliance.n.01")
        self.assertIsNone(self.taxonomy.get_parent("entity.n.01"))

    def test_is_leaf(self):
        self.assertTrue(self.taxonomy.is_leaf("stove.n.01"))
        self.assertFalse(self.taxonomy.is_leaf("kitchen_appliance.n.01"))
        self.assertFalse(self.taxonomy.is_leaf("entity.n.01"))

    def test_has_ability(self):
        self.assertTrue(self.taxonomy.has_ability("stove.n.01", "heatSource"))
        self.assertFalse(self.taxonomy.has_ability("stove.n.01", "burnable"))


if __name__ == '__main__':
    unittest.main()



import copy
import json
import pkgutil

import networkx as nx
import bddl

DEFAULT_HIERARCHY_FILE = pkgutil.get_data(
    bddl.__package__, "generated_data/output_hierarchy_properties.json")


class ObjectTaxonomy(object):
    def __init__(self, hierarchy_type="default"):
        hierarchy_file = DEFAULT_HIERARCHY_FILE 
        self.taxonomy = self._parse_taxonomy(hierarchy_file)

    @staticmethod
    def _parse_taxonomy(json_str):
        """
        Parse taxonomy from hierarchy JSON file.

        :param json_str: str containing JSON-encoded hierarchy.
        :return: networkx.DiGraph corresponding to object hierarchy tree with synsets as nodes and parent-child
            relationships as directed edges from parent to child.
        """
        json_obj = json.loads(json_str)

        taxonomy = nx.DiGraph()
        nodes = [(json_obj, None)]
        while len(nodes) > 0:
            next_nodes = []
            for node, parent in nodes:
                children_names = set()
                if 'children' in node:
                    for child in node['children']:
                        next_nodes.append((child, node))
                        children_names.add(child['name'])
                taxonomy.add_node(node['name'],
                                  categories=node.get('categories', []),
                                  substances=node.get('substances', []),
                                  abilities=node['abilities'])
                for child_name in children_names:
                    taxonomy.add_edge(node['name'], child_name)
            nodes = next_nodes
        return taxonomy

    def _get_synset_by_filter(self, filter_fn):
        """
        Gets a single synset matching the given filter function.

        :param filter_fn: Filter function that takes a single synset as input and returns a boolean (True to keep).
        :return: str corresponding to the matching synset, None if no match found.
        :raises: ValueError if more than one matching synset is found.
        """
        matched = [
            synset for synset in self.taxonomy.nodes if filter_fn(synset)]

        if not matched:
            return None
        elif len(matched) > 1:
            raise ValueError("Multiple synsets matched: %s" %
                             ", ".join(matched))

        return matched[0]
    
    def get_all_synsets(self):
        """
        Return all synsets in topology, in topological order (every synsets appears before its descendants)

        :return: list of str corresponding to synsets
        """
        return list(nx.topological_sort(self.taxonomy))

    def get_synset_from_category(self, category):
        """
        Get synset name corresponding to object category.

        :param category: object category to search for.
        :return: str containing matching synset.
        :raises ValueError if multiple matching synsets are found.
        """
        return self._get_synset_by_filter(lambda synset: category in self.get_categories(synset))

    def get_synset_from_substance(self, substance):
        """
        Get synset name corresponding to object category.

        :param substance: substance to search for.
        :return: str containing matching synset.
        :raises ValueError if multiple matching synsets are found.
        """
        return self._get_synset_by_filter(lambda synset: substance in self.get_substances(synset))

    def get_subtree_categories(self, synset):
        """
        Get the object categories matching the subtree of a given synset (by aggregating categories across all the leaf-level descendants).

        :param synset: synset to search
        :return: list of str corresponding to object categories
        """
        if self.is_leaf(synset):
            synsets = [synset]
        else:
            synsets = self.get_leaf_descendants(synset)
        all_categories = []
        for synset in synsets:
            all_categories += self.get_categories(synset)
        return all_categories
    
    def get_subtree_substances(self, synset):
        """
        Get the substances matching the subtree of a given synset (by aggregating substances across all the leaf-level descendants).

        :param synset: synset to search
        :return: list of str corresponding to substances
        """
        if self.is_leaf(synset):
            synsets = [synset]
        else:
            synsets = self.get_leaf_descendants(synset)
        all_substances = []
        for synset in synsets:
            all_substances += self.get_substances(synset)
        return all_substances

    def is_valid_synset(self, synset):
        """
        Check whether a given synset exists within the object taxonomy.

        :param synset: synset to search.
        :return: bool indicating if the synset exists in the taxonomy.
        """
        return self.taxonomy.has_node(synset)

    def get_descendants(self, synset):
        """
        Get the descendant synsets of a synset.

        :param synset: synset to search.
        :return: list of str corresponding to descendant synsets.
        """
        assert self.is_valid_synset(synset)
        return list(nx.algorithms.dag.descendants(self.taxonomy, synset))

    def get_leaf_descendants(self, synset):
        """
        Get the leaf descendant synsets of a synset, e.g. descendant synsets who are also leaf nodes.

        :param synset: synset to search.
        :return: list of str corresponding to leaf descendant synsets.
        """
        return [node for node in self.get_descendants(synset) if self.taxonomy.out_degree(node) == 0]

    def get_ancestors(self, synset):
        """
        Get the ancestor synsets of a synset.

        :param synset: synset to search.
        :return: list of str corresponding to ancestor synsets.
        """
        assert self.is_valid_synset(synset)
        return list(nx.algorithms.dag.ancestors(self.taxonomy, synset))

    def is_descendant(self, synset, potential_ancestor_synset):
        """
        Check whether a given synset is a descendant of another synset.

        :param synset: The synset being searched for as a descendant.
        :param potential_ancestor_synset: The synset being searched for as a parent.
        :return: bool indicating whether synset is a descendant of potential_ancestor_synset.
        """
        assert self.is_valid_synset(synset)
        assert self.is_valid_synset(potential_ancestor_synset)
        return synset in self.get_descendants(potential_ancestor_synset)

    def is_ancestor(self, synset, potential_descendant_synset):
        """
        Check whether a given synset is an ancestor of another synset.

        :param synset: The synset being searched for as an ancestor.
        :param potential_descendant_synset: The synset being searched for as a descendant.
        :return: bool indicating whether synset is an ancestor of potential_descendant_synset.
        """
        assert self.is_valid_synset(synset)
        assert self.is_valid_synset(potential_descendant_synset)
        return synset in self.get_ancestors(potential_descendant_synset)

    def get_abilities(self, synset):
        """
        Get the abilities of a given synset.

        :param synset: synset to search.
        :return: dict in the form of {ability: {param: value}} containing abilities and ability parameters.
        """
        assert self.is_valid_synset(synset), f"Invalid synset: {synset}"
        return copy.deepcopy(self.taxonomy.nodes[synset]['abilities'])

    def get_categories(self, synset):
        """
        Get the object categories matching a given synset.

        :param synset: synset to search.
        :return: list of str corresponding to object categories matching the synset.
        """
        assert self.is_valid_synset(synset)
        return list(self.taxonomy.nodes[synset]['categories'])
    
    def get_substances(self, synset):
        """
        Get the substances matching a given synset.

        :param synset: synset to search.
        :return: list of str corresponding to substances matching the synset.
        """
        assert self.is_valid_synset(synset)
        return list(self.taxonomy.nodes[synset]['substances'])

    def get_children(self, synset):
        """
        Get the immediate child synsets of a synset.

        :param synset: synset to search.
        :return: list of str corresponding to child synsets.
        """
        assert self.is_valid_synset(synset)
        return list(self.taxonomy.successors(synset))

    def get_parents(self, synset):
        """
        Get the immediate parent synsets of a synset.

        :param synset: synset to search.
        :return: list of str corresponding to parent synset
        """
        assert self.is_valid_synset(synset)

        return list(self.taxonomy.predecessors(synset))

    def is_leaf(self, synset):
        """
        Check whether a given synset is a leaf synset e.g. it has no descendants.

        :param synset: synset to search.
        :return: bool indicating if the synset is a leaf synset.
        """
        assert self.is_valid_synset(synset), "{} is not a valid synset".format(synset)
        return self.taxonomy.out_degree(synset) == 0

    def has_ability(self, synset, ability):
        """
        Check whether the given synset has the given ability.

        :param synset: synset to check.
        :param ability: Ability name to check.
        :return: bool indicating if the synset has the ability.
        """
        return ability in self.get_abilities(synset)


if __name__ == "__main__":
    object_taxonomy = ObjectTaxonomy()



"""
The program takes in UpWorker's .csv survey data, Manual Annotations (Canonical), and GPT Annotations (Canonical) and merges them.
This merged canonical is then run through the Programmatic Annotation system to create a Master Canonical. (Note: Master /= Final)
The Final step for the Master is to be run through propogate_by_intersection.py which has not been completed

Separately, once Master has been compiled, the two hierarchy generation functions will be run. The first function generates a hierarchy of all 
"""


from calendar import c
import json
import pathlib 
from nltk.corpus import wordnet as wn
import bddl.data_generation.prop_config as pcfg


CANONICAL_FN = pathlib.Path(__file__).parents[1] / "generated_data" / "syn_prop_annots_canonical.json"
PROP_TO_SYN_FILE = pathlib.Path(__file__).parents[1] / "generated_data" / "properties_to_synsets.json" # gives all applicable synsets for a given property
SYN_TO_DESC_FILE = pathlib.Path(__file__).parents[1] / "generated_data" / "synsets_to_descriptors.json" # gives states for a given synset

DESIRED_PROPS = set(pcfg.CROWD_PROP_NAMES + pcfg.GPT_PROP_NAMES + pcfg.INTERNAL_PROP_NAMES)

PROP_TO_DESC = { #added by hand from B1K Object States Google Doc found here https://docs.google.com/document/d/10g5A-ODF4POFN0SxIhiQIuDh5L1X_8KXriwKfaKWUFc/edit?usp=sharing as of May 12, 2022

    "breakable": "broken",
    "burnable": "burnt",
    "cleaningTool": None,
    "coldSource": None,
    "cookable": "cooked",
    "dustyable": "dusty",
    "freezable": "frozen",
    "heatSource": None,
    "liquid": None,
    "openable": "open",
    "perishable": "perished",
    "screwable": "screwed",
    "stainable": "stained",
    "sliceable": "sliced",
    "slicer": None,
    "diceable": "diced",
    "soakable": "soaked",
    "timeSetable": "time_set",
    "toggleable": "toggled_on",
    "waterSource": None,
    "foldable": "folded",
    "unfoldable": None,     # this property actually didn't even need to be annotated, because it's not used and it's just whatever's `foldable.` Or, empty and closed can be annotated, but similarly not put in desc list.
    "soakable": None,
    "wetable": "wet",
    "flammable": "on_fire",
    "assembleable": "assembled",
    "heatable" : "hot",
    "boilable": "boiling",
    "meltable" : "melted",
    "fireSource": None,
    "overcookable": "overcooked",
    "substance": None,
    "microPhysicalSubstance": None,
    "macroPhysicalSubstance": None,
    "visualSubstance": None,
    "deformable": None,
    "softBody": None,
    "cloth": None,
    "stickyable": "sticky",
    "grassyable": "grassy",
    "scratchable": "scratched",
    "tarnishable": "tarnished",
    "moldyable": "moldy",
    "rustable": "rusty",
    "wrinkleable": "wrinkly",
    "disinfectable": "disinfected",
    "mixable": None,
    "blendable": None,
    "rope": None,
    "fillable": None,
    "rigidBody": None,
    "nonSubstance": None,
    "particleRemover": None,
    "particleApplier": None,
    "particleSource": None,
    "particleSink": None,
    "needsOrientation": None,
    "waterCook": None,
    "nonDeformable": None,
    "nonSubstance": None,
    "sceneObject": None,
    "drapeable": None,
    "physicalSubstance": None,
    "mixingTool": None,
}

########### GET INPUT CANONICALS ###########

def get_annots_canonical(syn_prop_dict):
    canonical = {}
    for synset, prop_binaries in syn_prop_dict.items():
        synset_canonical = {}
        for prop, applies in prop_binaries.items():
            if prop in DESIRED_PROPS and bool(applies) and float(applies): 
                synset_canonical[prop] = {}
            elif prop == "objectType" and applies in DESIRED_PROPS:
                synset_canonical[applies] = {}
                if applies not in ["liquid", "visualSubstance", "macroPhysicalSubstance", "microPhysicalSubstance"]:
                    synset_canonical["nonSubstance"] = {}
        canonical[synset] = synset_canonical
    canonical_with_programmatic = add_programmatic_properties(canonical)
    return canonical_with_programmatic


# Removed original property annotation sources - human, GPT-3, and manual. If
#   we go back to those sources, we should write code to get the annotations 
#   into the unified Google sheet (that is then pulled here as a CSV) - better
#   way to handle things.


############ GETTING PROPERTIES #############

def add_programmatic_properties(synset_content): # runs programmatic addition over existing canonical input
    for synset in synset_content:
        if "liquid" in synset_content[synset]:
            synset_content[synset]["boilable"] = {}
        if ("visualSubstance" in synset_content[synset]) or ("microPhysicalSubstance" in synset_content[synset]) or ("macroPhysicalSubstance" in synset_content[synset]) or ("liquid" in synset_content[synset]):
            synset_content[synset]["substance"] = {}
        if ("microPhysicalSubstance" in synset_content[synset]) or ("macroPhysicalSubstance" in synset_content[synset]) or ("liquid" in synset_content[synset]):
            synset_content[synset]["physicalSubstance"] = {}
        if ("cloth" in synset_content[synset]) or ("rope" in synset_content[synset]) or ("softBody" in synset_content[synset]):
            synset_content[synset]["deformable"] = {}
        if ("cloth" in synset_content[synset]) or ("rope" in synset_content[synset]):
            synset_content[synset]["drapeable"] = {}
        if "nonSubstance" in synset_content[synset]: # non-substances are both wetable and mixable
                synset_content[synset].update({
                    "wetable": {},
                    "stickyable": {},
                    "dustyable": {},
                    "grassyable": {},
                    "scratchable": {},
                    "stainable": {},
                    "tarnishable": {},
                    "moldyable": {},
                    "rustable": {},
                    "wrinkleable": {},
                    "disinfectable": {},
                    "heatable": {}          # All nonsubstances are heatable. Substances are not heatable.
                })
        synset_content[synset]["freezable"] = {}    # all synsets are freezable

        # Hardcode in a place where the synset is seen, since not all scene synsets made it:
        if synset == "sink.n.01":
            synset_content[synset].update({"waterSource": {}})

    return synset_content


# Get properties_to_synsets.json
def make_properties_to_synsets(final_canonical):
    if type(final_canonical) == str: 
        with open(final_canonical, "r") as f:
            final_canonical = json.load(f)
    properties_to_synsets = {}
    for syn, props in final_canonical.items(): 
        for prop in props: 
            if prop not in properties_to_synsets:
                properties_to_synsets[prop] = []
            properties_to_synsets[prop].append(syn)
    return properties_to_synsets


def get_synset_descriptors(synsets_to_filtered_properties): # take in canonical format
    synsets_to_states = {} # declare an empty dictionary
    for synset_label, properties in synsets_to_filtered_properties.items():
        synsets_to_states[synset_label] = [[PROP_TO_DESC[prop]] * 2 for prop in properties if PROP_TO_DESC[prop] is not None] # if the property has an attached state, add its states
    return synsets_to_states


# API - only these should be used outside these script, and these should only be used outside this script

def create_get_save_annots_canonical(syn_prop_dict):
    print("Creating canonical annots file...")
    canonical = get_annots_canonical(syn_prop_dict)
    with open(CANONICAL_FN, "w") as f:
        json.dump(canonical, f, indent=4)
    print("Created and saved canonical annots file.")
    return canonical

def create_get_save_properties_to_synsets(propagated_canonical):
    print("Creating properties to synsets...")
    props_to_syns = make_properties_to_synsets(propagated_canonical)
    with open(PROP_TO_SYN_FILE, "w") as f:
        json.dump(props_to_syns, f, indent=4)
    print("Created and saved properties to synsets file.")
    return props_to_syns

def create_get_save_synsets_to_descriptors(propagated_canonical):
    print("Creating synsets to descriptors...")
    syn_to_desc = get_synset_descriptors(propagated_canonical)
    with open(SYN_TO_DESC_FILE, "w") as f:
        json.dump(syn_to_desc, f, indent=4)
    print("Created and saved synset to descriptor file.")
    return syn_to_desc



if __name__ == '__main__':
    pass


import os
import re
from bddl.condition_evaluation import (
    compile_state,
    create_scope,
    evaluate_state,
    get_ground_state_options,
)
from bddl.config import ACTIVITY_CONFIGS_PATH
from bddl.object_taxonomy import ObjectTaxonomy
from bddl.parsing import (
    gen_natural_language_condition,
    gen_natural_language_conditions,
    parse_domain,
    parse_problem,
)

INSTANCE_EXPR = re.compile(r"problem(\d+).bddl")


class Conditions(object):
    def __init__(self, behavior_activity, activity_definition, simulator_name, predefined_problem=None):
        """Object to store behavior activity content and compile conditions for checking and
            simulator use

        Args:
            behavior_activity (str): behavior activity being used
            activity_definition (int): specific definition of behavior_activity
            simulator_name (str): simulator that BEHAVIOR is being used with
            predefined_problem (str): a pre-defined problem that is not in the activity_definitions folder
        """
        self.behavior_activity = behavior_activity
        self.activity_definition = activity_definition
        domain_name, *__ = parse_domain(simulator_name)
        __, self.parsed_objects, self.parsed_initial_conditions, self.parsed_goal_conditions = parse_problem(
            self.behavior_activity, self.activity_definition, domain_name, predefined_problem=predefined_problem
        )


######## API ########


def get_object_scope(conds):
    """Create unpopulated object scope to populate for generating goal and
        ground goal conditions.

    Args:
        conds (Conditions): conditions for the particular activity and definition

    Returns:
        dict<str: None>: unpopulated scope with string keys to be mapped to
                            simulator object values
    """
    return create_scope(conds.parsed_objects)


def get_initial_conditions(conds, backend, scope, generate_ground_options=True):
    """Create compiled initial conditions that can be checked and sampled

    Args:
        conds (Conditions): conditions for the particular activity and definition

    Returns:
        list<bddl.condition_evaluation.HEAD>: compiled conditions if initial
                                                condition definition is not
                                                empty else None
    """
    if bool(conds.parsed_initial_conditions[0]):
        initial_conditions = compile_state(
            [cond for cond in conds.parsed_initial_conditions if cond[0] not in ["inroom"]],
            backend,
            scope=scope,
            object_map=conds.parsed_objects,
            generate_ground_options=generate_ground_options
        )
        return initial_conditions


def get_goal_conditions(conds, backend, scope, generate_ground_options=True):
    """Create compiled goal conditions with a populated object scope for checking

    Args:
        conds (Conditions): conditions for the particular activity and definition
        populated_object_scope (dict<str: simulator object>): scope mapping object
                                                                terms in BDDL to
                                                                simulator objects

    Returns:
        list<bddl.condition_evaluation.HEAD>: compiled conditions if goal condition
                                                definition is not empty else None
    """
    if bool(conds.parsed_goal_conditions[0]):
        goal_conditions = compile_state(
            conds.parsed_goal_conditions, 
            backend, 
            scope=scope, 
            object_map=conds.parsed_objects,
            generate_ground_options=generate_ground_options
        )
        return goal_conditions


def get_ground_goal_state_options(conds, backend, scope, goal_conditions):
    """Create compiled ground solutions to goal state with a populated object scope
        for checking progress on specific solutions

    Args:
        conds (Conditions): conditions for the particular activity and definition
        populated_object_scope (dict<str: simulator object>): scope mapping object
                                                                terms in BDDL to
                                                                simulator objects

    Returns:
        list<bddl.condition_evaluation.HEAD>: compiled goal solutions

    Raises:
        AssertionError if there are no ground solutions
    """
    ground_goal_state_options = get_ground_state_options(
        goal_conditions, backend, scope=scope, object_map=conds.parsed_objects
    )
    assert len(ground_goal_state_options) > 0
    return ground_goal_state_options


def evaluate_goal_conditions(goal_conditions):
    """Evaluate compiled goal state to see if current simulator state has been met

    Args:
        goal_conditions (list<bddl.condition_evaluation.HEAD>): list of compiled
                                                                goal conditions with
                                                                populated scope

    Returns:
        bool, dict<str: list<int>>: [description]
    """
    return evaluate_state(goal_conditions)


def get_natural_initial_conditions(conds):
    """Return natural language translation of init of given conditions

    Args:
        conditions (list): conditions being translated

    Returns:
        list<str>: natural language translations, one per condition in conditions
    """
    return gen_natural_language_conditions(conds.parsed_initial_conditions)


def get_natural_goal_conditions(conds):
    """Return natural language translation of goal of given conditions

    Args:
        conditions (list): conditions being translated

    Returns:
        list<str>: natural language translations, one per condition in conditions
    """
    return gen_natural_language_conditions(conds.parsed_goal_conditions)


def get_all_activities():
    """Return a list of all activities included in this version of BDDL.
        
    Returns:
        list<str>: list containing the name of each included activity
    """
    return [x for x in os.listdir(ACTIVITY_CONFIGS_PATH) if os.path.isdir(os.path.join(ACTIVITY_CONFIGS_PATH, x))]


def get_instance_count(act):
    """Return the number of instances of a given activity that are included in this version of BDDL.
    
    Args:
        act (str): name of the activity to check
        
    Returns:
        int: number of instances of the given activity
    """
    problem_files = [INSTANCE_EXPR.fullmatch(x) for x in os.listdir(os.path.join(ACTIVITY_CONFIGS_PATH, act))]
    ids = set(int(x.group(1)) for x in problem_files if x is not None)
    assert ids == set(range(len(ids))), f"Non-contiguous instance IDs found for problem {act}"
    return len(ids)


def get_reward(ground_goal_state_options): 
    """Return reward given ground goal state options.
       Reward formulated as max(<percent literals that are satisfied in the option> for option in ground_goal_state_options)

    Args: 
        ground_goal_state_options (list<list<HEAD>>): list of compiled ground goal state options

    Returns: 
        float: reward
    """
    return max(len(evaluate_state(option)[-1]["satisfied"]) / float(len(option)) for option in ground_goal_state_options)



###### COMBINATORICS UTILS ######

import itertools
from bddl.config import GROUND_GOALS_MAX_OPTIONS, GROUND_GOALS_MAX_PERMUTATIONS


def truncated_product(*sequences, max_options=GROUND_GOALS_MAX_OPTIONS):
    """Breadth-first search cartesian product 
       Source: https://stackoverflow.com/questions/42288203/generate-itertools-product-in-different-order

    :yields (tuple): elements of cartesian product 
    """
    # sequences = tuple(tuple(seq) for seqin sequences)
    counter = 0

    def partitions(n, k):
        for c in itertools.combinations(range(n + k - 1), k - 1):
            yield (b - a - 1 for a, b in zip((-1,) + c, c + (n + k -1,)))

    max_position = [len(i) - 1 for i in sequences]
    for i in range(sum(max_position)):
        if counter >= max_options:
            break 
        for positions in partitions(i, len(sequences)):
            try:
                if counter < max_options:
                    counter += 1
                    yield tuple(map(lambda seq, pos: seq[pos], sequences, positions))
                else: 
                    break
            except IndexError:
                continue
    # print("next try:", tuple(map(lambda seq, pos: seq[pos], sequences, max_position)))
    if counter < max_options:
        counter += 1
        yield tuple(map(lambda seq, pos: seq[pos], sequences, max_position))


def truncated_permutations(iterable, r=None, max_permutations=GROUND_GOALS_MAX_PERMUTATIONS):
    """Adapted from https://docs.python.org/3/library/itertools.html#itertools.permutations
    """
    pool = tuple(iterable)
    n = len(pool)
    r = n if r is None else r 
    if r > n:
        return 
    indices = list(range(n))
    cycles = list(range(n, n - r, -1))

    counter = 0
    if counter < max_permutations:
        counter += 1
        yield tuple(pool[i] for i in indices[:r])
    while n:
        for i in reversed(range(r)):
            cycles[i] -= 1
            if cycles[i] == 0:
                indices[i:] = indices[i + 1:] + indices[i : i + 1]
                cycles[i] = n - i
            else:
                j = cycles[i]
                indices[i], indices[-j] = indices[-j], indices[i]
                if counter < max_permutations:
                    counter += 1
                    yield tuple(pool[i] for i in indices[:r])
                else:
                    return 
                break
        else:
            return


########## CUSTOM ERRORS ############

class UncontrolledCategoryError(Exception):
    def __init__(self, malformed_cat):
        self.malformed_cat = malformed_cat

class UnsupportedPredicateError(Exception):
    def __init__(self, predicate):
        self.predicate = predicate


if __name__ == "__main__":
    import pprint
    sample_small = [[[['ap1', 'tab1']], [['ap1', 'tab2']]], [[['ap2', 'tab1']], [['ap2', 'tab2']]], [[['ap3', 'tab1']], [['ap3', 'tab2']]]]
    # pprint.pprint(list(truncated_product(*sample_small)))
    # print(list(truncated_permutations(range(10), 3)))
    list(truncated_permutations(range(10), 3))



from bddl.backend_abc import BDDLBackend

_AVAILABLE_BACKENDS = ["iGibson"]
_backend = None


def set_backend(backend_name):
    global _backend
    if backend_name == "iGibson":
        from igibson.tasks.bddl_backend import IGibsonBDDLBackend
        _backend = IGibsonBDDLBackend()
    else:
        raise ValueError("Invalid backend. Currently supported backends: %s." % ", ".join(_AVAILABLE_BACKENDS))

    if not isinstance(_backend, BDDLBackend):
        raise ValueError("Backends must implement bddl.backend_abc.BDDLBackend.")



def get_backend():
    if _backend is None:
        raise ValueError(
            "Before calling bddl functions, a backend must be set using bddl.set_backend(backend_name). "
            "Available backend names: %s." % ", ".join(_AVAILABLE_BACKENDS))

    return _backend


CROWD_PROP_NAMES = [
    "breakable",
    "fillable",
    "flammable",
    "openable", 
    "toggleable",
]

GPT_PROP_NAMES = [
    "cookable", 
    "heatSource",
    "coldSource",
    # "fireSource",       # deprecated
    "sliceable",
    "slicer",
    # "soakable",         # deprecated
    # "cleaningTool",     # deprecated
]

INTERNAL_PROP_NAMES = [
    "liquid",
    "visualSubstance",
    "microPhysicalSubstance",
    "macroPhysicalSubstance",
    "cloth",
    "rope",
    "softBody",
    "rigidBody",
    "assembleable",
    "meltable",
    "particleRemover",
    "particleApplier",
    "particleSource",
    "particleSink",
    "needsOrientation",
    "waterCook",
    "mixingTool",
    "sceneObject",
    "diceable",
]



import json 
import csv
import pandas as pd
import pathlib
import os
from collections import Counter
import re

from bddl.data_generation.tm_submap_params import TM_SUBMAPS_TO_PARAMS

SHEETS_DIR = pathlib.Path(__file__).parents[1] / "generated_data" / "transition_map" / "tm_raw_data"
JSONS_DIR = pathlib.Path(__file__).parents[1] / "generated_data" / "transition_map" / "tm_jsons"

OBJECT_CAT_AND_INST_RE = r"[A-Za-z-_]+\.n\.[0-9]+"


def parse_conditions_entry(unparsed_conditions):
    # print(f"Parsing: {unparsed_conditions}")
    if unparsed_conditions.isnumeric():
        always_true = bool(int(unparsed_conditions))
        conditions = [] if always_true else None
    else:
        conditions = unparsed_conditions.lower().split(" or ")
    return conditions


def sheet_to_json(submap):
    params = TM_SUBMAPS_TO_PARAMS[submap]
    raw_data = pd.read_csv(os.path.join(SHEETS_DIR, submap + ".csv"))[params].to_json(orient="records")
    data = json.loads(raw_data)
    reformatted_data = []
    for rule in data:
        reformatted_rule = {}
        for param, value in rule.items():
            if TM_SUBMAPS_TO_PARAMS[submap][param]["type"] == "synset" and value is not None:
                value = value.split(" and ")
                value = Counter([re.match(OBJECT_CAT_AND_INST_RE, atom).group() for atom in value])
            elif TM_SUBMAPS_TO_PARAMS[submap][param]["type"] == "atom" and value is not None:
                value = value.split(" and ")
                unary_atoms = []
                binary_atoms = []
                for atom in value: 
                    elements = atom.split(" ")
                    assert len(elements) in range(2, 5), f"Malformed atom {atom}"
                    if len(elements) == 3:
                        if elements[0] == "not":
                            unary_atoms.append(elements)
                        else:
                            binary_atoms.append(elements)
                    elif len(elements) == 2:
                        unary_atoms.append(elements)
                    else:
                        binary_atoms.append(elements)
                reformatted_atoms = {}
                for atom in unary_atoms: 
                    synset = re.match(OBJECT_CAT_AND_INST_RE, atom[-1]).group()
                    if synset in reformatted_rule:
                        reformatted_atoms[synset].append((atom[-2], atom[0] != "not"))
                    else:
                        reformatted_atoms[synset] = [(atom[-2], atom[0] != "not")]
                for atom in binary_atoms:
                    synset1, synset2 = re.match(OBJECT_CAT_AND_INST_RE, atom[-2]).group(), re.match(OBJECT_CAT_AND_INST_RE, atom[-1]).group()
                    if f"{synset1},{synset2}" in reformatted_rule:
                        reformatted_atoms[f"{synset1},{synset2}"].append((atom[-3], atom[0] != "not"))
                    else:
                        reformatted_atoms[f"{synset1},{synset2}"] = [(atom[-3], atom[0] != "not")]
                value = reformatted_atoms
            elif TM_SUBMAPS_TO_PARAMS[submap][param]["type"] == "string":
                value = value
            elif value is None:
                value = None
            else: 
                raise ValueError(f"Unhandled parameter type {TM_SUBMAPS_TO_PARAMS[submap][param]['type']}")
            reformatted_rule[param] = value
        reformatted_data.append(reformatted_rule)

    with open(os.path.join(JSONS_DIR, submap + ".json"), "w") as f:
        json.dump(reformatted_data, f, indent=4)


def sheet_to_json_washer():
    records = []
    with open(os.path.join(SHEETS_DIR, "washer.csv")) as f:
        reader = csv.DictReader(f)
        for row in reader:
            records.append(row)
    assert len(records) == 1
    record = records[0]

    remover_kwargs = dict()

    # Iterate through all the columns headed by a substance, in no particular order since their ultimate location is a dict
    for dirtiness_substance_synset in [key for key in record if re.match(OBJECT_CAT_AND_INST_RE, key) is not None]:
        conditions = parse_conditions_entry(record[dirtiness_substance_synset])
        remover_kwargs[dirtiness_substance_synset] = conditions

    with open(os.path.join(JSONS_DIR, "washer.json"), "w") as f:
        json.dump(remover_kwargs, f, indent=4)

def create_save_explicit_transition_rules():
    print("Creating explicit transition rule jsons...")
    for submap in TM_SUBMAPS_TO_PARAMS:
        sheet_to_json(submap)

    sheet_to_json_washer()
    print("Created and saved explicit transition rule jsons.")


if __name__ == "__main__":
    for submap in TM_SUBMAPS_TO_PARAMS:
        sheet_to_json(submap)


from abc import abstractmethod, ABCMeta

from future.utils import with_metaclass


class BDDLBackend(with_metaclass(ABCMeta)):
    @abstractmethod
    def get_predicate_class(self, predicate_name):
        """Given predicate_name, return an implementation of bddl.logic_base.AtomicFormula or subclasses."""
        pass